# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

import copy
import warnings

import numpy as np
from abc import ABC, abstractmethod

from grid2op.Observation import BaseObservation
from grid2op.Exceptions import PlotError
from grid2op.PlotGrid.LayoutUtil import layout_obs_sub_load_and_gen
from grid2op.PlotGrid.PlotUtil import PlotUtil as pltu
from grid2op.dtypes import dt_float, dt_int


class BasePlot(ABC):
    """
    INTERNAL

    .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\

    Abstract interface to plot the state of the powergrid

    Implement the interface with a plotting library to generate drawings 

    Attributes
    -----------

    observation_space: ``grid2op.Observation.ObservationSpace``
        The observation space used.

    width: ``int`` Width of the drawing

    height: ``int`` Height of the drawing

    grid_layout: ``dict`` A grid layout dict to use

    """

    def __init__(
        self,
        observation_space,
        width=800,
        height=600,
        scale=2000.0,
        grid_layout=None,
        parallel_spacing=3.0,
    ):

        self.observation_space = observation_space
        self.width = width
        self.height = height
        self.scale = scale
        self._parallel_spacing = parallel_spacing

        self._info_to_units = {"rho": "%", "a": "A", "p": "MW", "v": "kV"}
        self._lines_info = ["rho", "a", "p", "v", None]
        self._loads_info = ["p", "v", None]
        self._gens_info = ["p", "v", None]

        self._grid_layout = self.compute_grid_layout(observation_space, grid_layout)

        # Augment observation_space with dummy observation data
        # so we can use it as an observation for plotting just the layout or custom infos
        self.observation_space.topo_vect = np.ones(
            self.observation_space.dim_topo, dtype=dt_int
        )
        self.observation_space.line_status = np.full(
            self.observation_space.n_line, True
        )
        self.observation_space.rho = np.full(self.observation_space.n_line, 0.0)
        self.observation_space.p_or = np.ones(self.observation_space.n_line)

        # TODO storage: display the storage units too
        # TODO storage doc (yes also the documentation)

    @abstractmethod
    def create_figure(self):
        """
        Creates a new figure to draw into.
        Depending on the library can also be called Plot, canvas, screen ..
        """
        pass

    @abstractmethod
    def clear_figure(self, figure):
        """
        Clears a figure
        Depending on the library can also be called Plot, canvas, screen ..
        """
        pass

    @abstractmethod
    def convert_figure_to_numpy_HWC(self, figure):
        """
        Given a figure as returned by `BasePlot.create_figure`
        Convert it to a numpy array of dtype uint8
        and data layed out in the HWC format
        """
        pass

    def compute_grid_layout(self, observation_space, grid_layout=None):
        """
        Compute the grid layout from the observation space

        This should return a native python ``dict``
        in the same format as observation_space.grid_layout :

        .. code-block:: python

            {
              "substation1_name": [x_coord, y_coord],
              "substation2_name": [x_coord, y_coord],
              [...],
              "load1_name": [x_coord, y_coord],
              [...],
              "gen1_name": [x_coord, y_coord],
              [...]
            }

        Note that is must contain at least the positions for the substations.
        The loads and generators will be skipped if missing.

        By default, if `grid_layout` is provided this is returned,
        otherwise returns observation_space.grid_layout

        Parameters
        ----------

        observation_space: ``grid2op.Observation.ObservationSpace``
             The observation space of the environment

        grid_layout: ``dict`` or ``None``
             A dictionary containing the coordinates for each substation.
        """
        # We need an intial layout to work with
        use_grid_layout = None
        if grid_layout is not None:
            use_grid_layout = grid_layout
        elif observation_space.grid_layout is not None:
            use_grid_layout = observation_space.grid_layout
        else:
            raise PlotError("No grid layout provided for plotting")

        # Compute loads and gens positions using a default implementation
        observation_space.grid_layout = use_grid_layout
        return layout_obs_sub_load_and_gen(
            observation_space, scale=self.scale, use_initial=True
        )

    @abstractmethod
    def draw_substation(self, figure, observation, sub_id, sub_name, pos_x, pos_y):
        """
        Draws a substation into the figure

        Parameters
        ----------

        figure: :object: Figure to draw to.
        This is the object returned by create_figure

        observation: :grid2op.Observation.BaseObservation:
        Current state of the grid being drawn

        sub_id: :int: Id of the substation, Index in the observation

        sub_name: :str: Name of the substation

        pos_x: :int: x position from the layout

        pos_y: :int: y position from the layout
        """
        pass

    def update_substation(self, figure, observation, sub_id, sub_name, pos_x, pos_y):
        """
        INTERNAL

        .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\

        Update a substation into the figure
        """
        pass

    @abstractmethod
    def draw_load(
        self,
        figure,
        observation,
        load_name,
        load_id,
        load_bus,
        load_value,
        load_unit,
        pos_x,
        pos_y,
        sub_x,
        sub_y,
    ):
        """
        Draws a load into the figure

        Parameters
        ----------

        figure: :object: Figure to draw to.
        This is the object returned by create_figure

        observation: :grid2op.Observation.BaseObservation:
        Current state of the grid being drawn

        load_name: ``str`` Name of the load

        load_id: ``int`` Id of the load, Index in the observation

        load_bus: ``int`` Id of bus the load is connected to.

        load_value: ``float`` An informative value of the load current state

        load_unit: ``str`` The unit of the `load_value` argument as a string

        pos_x: ``int`` x position from the layout

        pos_y: ``int`` y position from the layout

        sub_x: ``int`` x position of the connected substation from the layout

        sub_y: ``int`` y position of the connected substation from the layout
        """
        pass

    def update_load(
        self,
        figure,
        observation,
        load_name,
        load_id,
        load_bus,
        load_value,
        load_unit,
        pos_x,
        pos_y,
        sub_x,
        sub_y,
    ):
        """
        INTERNAL

        .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\

        Update a load into the figure
        """
        pass

    @abstractmethod
    def draw_gen(
        self,
        figure,
        observation,
        gen_name,
        gen_id,
        gen_bus,
        gen_value,
        gen_unit,
        pos_x,
        pos_y,
        sub_x,
        sub_y,
    ):
        """
        Draws a generator into the figure

        Parameters
        ----------

        figure: :object: Figure to draw to.
        This is the object returned by create_figure

        observation: `grid2op.Observation.BaseObservation`
        Current state of the grid being drawn

        gen_name: ``str`` Name of the load

        gen_id: ``int`` Id of the generator, Index in the observation

        gen_bus: ``int`` Bus id the generator is connected to

        gen_value: ``float``
        An informative value of the generator current state

        gen_unit: ``str`` The unit of the `gen_value` argument as a string

        pos_x: ``int`` x position from the layout

        pos_y: ``int`` y position from the layout

        sub_x: ``int`` x position of the connected substation from the layout

        sub_y: ``int`` y position of the connected substation from the layout
        """
        pass

    def update_gen(
        self,
        figure,
        observation,
        gen_name,
        gen_id,
        gen_bus,
        gen_value,
        gen_unit,
        pos_x,
        pos_y,
        sub_x,
        sub_y,
    ):
        """
        INTERNAL

        .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\

        Updates a generator into the figure
        """
        pass

    @abstractmethod
    def draw_powerline(
        self,
        figure,
        observation,
        line_id,
        line_name,
        connected,
        line_value,
        line_unit,
        or_bus,
        pos_or_x,
        pos_or_y,
        ex_bus,
        pos_ex_x,
        pos_ex_y,
    ):
        """
        Draws a powerline into the figure

        Parameters
        ----------

        figure: ``object`` Figure to draw to.
        This is the object returned by `create_figure`

        observation: ``grid2op.Observation.BaseObservation``
        Current state of the grid being drawn

        line_id: ``int`` Id of the powerline, index in the observation

        line_name: ``str`` Name of the powerline

        connected: ``bool`` Is the line connected ?

        line_value: ``float`` An informative value of the line current state

        line_unit: ``str`` The unit of the `line_value` argument as a string

        or_bus: ``int`` Bus the powerline origin is connected to

        pos_or_x: ``int`` Powerline origin x position from the layout

        pos_or_y: ``int`` Powerline origin y position from the layout

        ex_bus: ``int`` Bus the powerline extremity is connected to

        pos_ex_x: ``int`` Powerline extremity x position from the layout

        pos_ex_y: ``int`` Powerline extremity y position from the layout
        """
        pass

    def update_powerline(
        self,
        figure,
        observation,
        line_id,
        line_name,
        connected,
        line_value,
        line_unit,
        or_bus,
        pos_or_x,
        pos_or_y,
        ex_bus,
        pos_ex_x,
        pos_ex_y,
    ):
        """
        INTERNAL

        .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\

        Draws a powerline into the figure
        """
        pass

    @abstractmethod
    def draw_storage(
        self,
        figure,
        observation,
        storage_name,
        storage_id,
        storage_bus,
        storage_value,
        storage_unit,
        pos_x,
        pos_y,
        sub_x,
        sub_y,
    ):
        """
        Draws a storage unit into the figure

        Parameters
        ----------

        figure: :object: Figure to draw to.
        This is the object returned by create_figure

        observation: :grid2op.Observation.BaseObservation:
        Current state of the grid being drawn

        storage_name: ``str`` Name of the load

        storage_id: ``int`` Id of the load, Index in the observation

        storage_bus: ``int`` Id of bus the load is connected to.

        storage_value: ``float`` An informative value of the load current state

        storage_unit: ``str`` The unit of the `load_value` argument as a string

        pos_x: ``int`` x position from the layout

        pos_y: ``int`` y position from the layout

        sub_x: ``int`` x position of the connected substation from the layout

        sub_y: ``int`` y position of the connected substation from the layout
        """
        pass

    def update_storage(
        self,
        figure,
        observation,
        storage_name,
        storage_id,
        storage_bus,
        storage_value,
        storage_unit,
        pos_x,
        pos_y,
        sub_x,
        sub_y,
    ):
        """
        INTERNAL

        .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\

        Update a storage unit into the figure
        """
        pass

    @abstractmethod
    def draw_legend(self, figure, observation):
        """
        Setup the legend for the given figure.

        Parameters
        ----------
        figure: ``object`` Figure to draw to.
        This is the object returned by `create_figure`

        observation: ``grid2op.Observation.BaseObservation``
        Current state of the grid being drawn
        """
        pass

    def update_legend(self, figure, observation):
        """
        INTERNAL

        .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\

        Updates the legend for the given figure.
        """
        pass

    def plot_postprocess(self, figure, observation, is_update):
        """
        Some implementations may need post-processing.
        This is called at the end of plot.
        """
        pass

    def _plot_subs(self, figure, observation, redraw):
        draw_fn = self.draw_substation
        if not redraw:
            draw_fn = self.update_substation

        for sub_idx in range(observation.n_sub):
            sub_name = observation.name_sub[sub_idx]
            sub_x = self._grid_layout[sub_name][0]
            sub_y = self._grid_layout[sub_name][1]
            draw_fn(figure, observation, sub_idx, sub_name, sub_x, sub_y)

    def _aux_draw_elements(
        self,
        figure,
        observation,
        load_values,
        load_unit,
        draw_fn,
        el_names,
        el_to_subid,
        el_pos_topo_vect,
    ):
        """
        generic method to loop through all elements of a given type and call the draw function
        on them
        """
        topo = observation.topo_vect
        topo_pos = el_pos_topo_vect
        for stor_idx, stor_name in enumerate(el_names):
            if stor_name not in self._grid_layout:
                continue
            load_value = None
            if load_values is not None:
                if load_values[stor_idx] is not None:
                    load_value = np.round(float(load_values[stor_idx]), 2)
                else:
                    load_value = None
            sto_x = self._grid_layout[stor_name][0]
            sto_y = self._grid_layout[stor_name][1]
            sto_subid = el_to_subid[stor_idx]
            subname = observation.name_sub[sto_subid]
            sto_bus = topo[topo_pos[stor_idx]]
            sto_bus = sto_bus if sto_bus > 0 else 0
            sub_x = self._grid_layout[subname][0]
            sub_y = self._grid_layout[subname][1]
            draw_fn(
                figure,
                observation,
                stor_idx,
                stor_name,
                sto_bus,
                load_value,
                load_unit,
                sto_x,
                sto_y,
                sub_x,
                sub_y,
            )

    def _plot_loads(self, figure, observation, load_values, load_unit, redraw):
        draw_fn = self.draw_load
        if not redraw:
            draw_fn = self.update_load

        self._aux_draw_elements(
            figure,
            observation,
            load_values,
            load_unit,
            draw_fn,
            observation.name_load,
            observation.load_to_subid,
            observation.load_pos_topo_vect,
        )

    def _plot_storages(self, figure, observation, storage_values, storage_unit, redraw):
        if observation.n_storage == 0:
            return
        draw_fn = self.draw_storage
        if not redraw:
            draw_fn = self.update_storage
        self._aux_draw_elements(
            figure,
            observation,
            storage_values,
            storage_unit,
            draw_fn,
            observation.name_storage,
            observation.storage_to_subid,
            observation.storage_pos_topo_vect,
        )

    def _plot_gens(self, figure, observation, gen_values, gen_unit, redraw):
        draw_fn = self.draw_gen
        if not redraw:
            draw_fn = self.update_gen

        self._aux_draw_elements(
            figure,
            observation,
            gen_values,
            gen_unit,
            draw_fn,
            observation.name_gen,
            observation.gen_to_subid,
            observation.gen_pos_topo_vect,
        )

    def _plot_lines(self, figure, observation, line_values, line_unit, redraw):
        draw_fn = self.draw_powerline
        if not redraw:
            draw_fn = self.update_powerline

        topo = observation.topo_vect
        line_or_pos = observation.line_or_pos_topo_vect
        line_ex_pos = observation.line_ex_pos_topo_vect

        for line_idx in range(observation.n_line):
            line_or_sub = observation.line_or_to_subid[line_idx]
            line_or_sub_name = observation.name_sub[line_or_sub]
            line_ex_sub = observation.line_ex_to_subid[line_idx]
            line_ex_sub_name = observation.name_sub[line_ex_sub]
            line_name = observation.name_line[line_idx]
            line_status = True
            line_status = observation.line_status[line_idx]
            line_value = None
            if line_values is not None:
                lv = line_values[line_idx]
                if isinstance(lv, (float, dt_float)):
                    line_value = np.round(float(lv), 2)
                elif isinstance(lv, (int, dt_int)):
                    line_value = int(lv)
                else:
                    line_value = lv

            line_or_bus = topo[line_or_pos[line_idx]]
            line_or_bus = line_or_bus if line_or_bus > 0 else 0
            line_or_x = self._grid_layout[line_or_sub_name][0]
            line_or_y = self._grid_layout[line_or_sub_name][1]
            line_ex_bus = topo[line_ex_pos[line_idx]]
            line_ex_bus = line_ex_bus if line_ex_bus > 0 else 0
            line_ex_x = self._grid_layout[line_ex_sub_name][0]
            line_ex_y = self._grid_layout[line_ex_sub_name][1]

            # Special case for parralel lines
            tmp = self.observation_space.get_lines_id(
                from_=line_or_sub, to_=line_ex_sub
            )
            if len(tmp) > 1:
                ox, oy = pltu.orth_norm_from_points(
                    line_or_x, line_or_y, line_ex_x, line_ex_y
                )
                if line_idx == tmp[0]:
                    line_or_x += ox * self._parallel_spacing
                    line_or_y += oy * self._parallel_spacing
                    line_ex_x += ox * self._parallel_spacing
                    line_ex_y += oy * self._parallel_spacing
                else:
                    line_or_x -= ox * self._parallel_spacing
                    line_or_y -= oy * self._parallel_spacing
                    line_ex_x -= ox * self._parallel_spacing
                    line_ex_y -= oy * self._parallel_spacing

            draw_fn(
                figure,
                observation,
                line_idx,
                line_name,
                line_status,
                line_value,
                line_unit,
                line_or_bus,
                line_or_x,
                line_or_y,
                line_ex_bus,
                line_ex_x,
                line_ex_y,
            )

    def _plot_legend(self, fig, observation, redraw):
        draw_fn = self.draw_legend
        if not redraw:
            draw_fn = self.update_legend

        draw_fn(fig, observation)

    def plot_layout(self):
        """
        This function plot the layout of the grid, as well as the object. You will see the name of each elements and
        their id.
        """
        return self.plot_info(
            observation=self.observation_space, figure=None, redraw=True
        )

    def plot_obs(
        self,
        observation,
        figure=None,
        redraw=True,
        line_info="rho",
        load_info="p",
        gen_info="p",
        storage_info="p",
    ):
        """
        Plot an observation.

        Parameters
        ----------
        observation: :class:`grid2op.Observation.BaseObservation`
            The observation to plot

        figure: ``object``
            The figure on which to plot the observation.
            If figure is ``None``, a new figure is created.

        line_info: ``str``
            One of "rho", "a", or "p" or "v"
            The information that will be plotted on the powerline.
            By default "rho".
            All flow are taken "origin" side.

        load_info: ``str``
            One of "p" or "v" the information displayed on the load.
            (default to "p").

        gen_info: ``str``
            One of "p" or "v" the information displayed on the generators
            (default to "p").

        storage_info: ``str``
            One of "p" or None the information displayed on the generators
            (default to "p").

        Returns
        -------
        res: ``object``
            The figure updated with the data from the new observation.
        """
        # TODO add gen_info=dispatch and gen_info=target_dispatch
        # TODO add line_info=cooldown

        # Start by checking arguments are valid
        if not isinstance(observation, BaseObservation):
            err_msg = (
                "Observation is not a derived type of "
                "grid2op.Observation.BaseObservation"
            )
            raise PlotError(err_msg)
        if line_info not in self._lines_info:
            err_msg = (
                'Impossible to plot line info "{}" for line.' " Possible values are {}"
            )
            raise PlotError(err_msg.format(line_info, str(self._lines_info)))
        if load_info not in self._loads_info:
            err_msg = (
                'Impossible to plot load info "{}" for line.' " Possible values are {}"
            )
            raise PlotError(err_msg.format(load_info, str(self._loads_info)))
        if gen_info not in self._gens_info:
            err_msg = (
                'Impossible to plot gen info "{}" for line.' " Possible values are {}"
            )
            raise PlotError(err_msg.format(gen_info, str(self._gens_info)))

        line_values = None
        line_unit = ""
        if line_info is not None:
            line_unit = self._info_to_units[line_info]
        if line_info == "rho":
            line_values = observation.rho * 100.0
        elif line_info == "p":
            line_values = observation.p_or
        elif line_info == "a":
            line_values = observation.a_or
        elif line_info == "v":
            line_values = observation.v_or
        elif line_info is None:
            pass
        else:
            raise PlotError(
                f'Impossible to understand the keyword argument "{line_info}" '
                f'provided as "line_info"'
            )

        load_values = None
        load_unit = ""
        if load_info is not None:
            load_unit = self._info_to_units[load_info]
        if load_info == "p":
            load_values = copy.copy(observation.load_p) * -1.0
        elif load_info == "v":
            load_values = observation.load_v
        elif load_info is None:
            pass
        else:
            raise PlotError(
                f'Impossible to understand the keyword argument "{load_info}" '
                f'provided as "load_info"'
            )

        gen_values = None
        gen_unit = ""
        if gen_info is not None:
            gen_unit = self._info_to_units[gen_info]
        if gen_info == "p":
            gen_values = observation.prod_p
        elif gen_info == "v":
            gen_values = observation.prod_v
        elif gen_info is None:
            pass
        else:
            raise PlotError(
                f'Impossible to understand the keyword argument "{gen_info}" '
                f'provided as "gen_info"'
            )

        storage_values = None
        storage_unit = ""
        if storage_info is not None:
            storage_unit = self._info_to_units[storage_info]
        if storage_info == "p":
            storage_values = -1.0 * observation.storage_power
        elif storage_info is None:
            pass
        else:
            raise PlotError(
                f'Impossible to understand the keyword argument "{storage_info}" '
                f'provided as "storage_info"'
            )

        return self.plot_info(
            observation=observation,
            figure=figure,
            redraw=redraw,
            line_values=line_values,
            line_unit=line_unit,
            load_values=load_values,
            load_unit=load_unit,
            gen_values=gen_values,
            gen_unit=gen_unit,
            storage_values=storage_values,
            storage_unit=storage_unit,
        )

    def plot_info(
        self,
        figure=None,
        redraw=True,
        line_values=None,
        line_unit="",
        load_values=None,
        load_unit="",
        storage_values=None,
        storage_unit="",
        gen_values=None,
        gen_unit="",
        observation=None,
        coloring=None,
    ):
        """
        Plot an observation with custom values

        Parameters
        ----------
        figure: ``object``
            The figure on which to plot the observation.
            If figure is ``None`` a new figure is created.

        line_values: ``list``
            information to be displayed for the powerlines
            [must have the same size as observation.n_line  and convertible to float]

        line_unit: ``str``
            Unit string for the :line_values: argument, displayed after the line value

        load_values: ``list``
            information to display for the loads
            [must have the same size as observation.n_load and convertible to float]

        load_unit: ``str``
            Unit string for the :load_values: argument, displayed after the load value

        storage_values: ``list``
            information to display for the storage units
            [must have the same size as observation.n_storage and convertible to float]

        storage_unit: ``str``
            Unit string for the :storage_values: argument, displayed after the storage value

        gen_values: ``list``
            information to display in the generators
            [must have the same size as observation.n_gen and convertible to float]

        gen_unit: ``str``
            Unit string for the :gen_values: argument, displayed after the generator value

        observation: :class:`grid2op.Observation.BaseObservation`
            An observation to plot, can be None if no values are drawn from the observation

        coloring:
            ``None`` for no special coloring, or "line" to color the powerline based on the value
            ("gen" and "load" coming soon)

        Examples
        --------

        More examples on how to use this function is given in the "8_PlottingCapabilities.ipynb" notebook.

        The basic concept is:

        .. code-block:: python

            import grid2op
            from grid2op.PlotGrid import PlotMatplot
            env = grid2op.make()
            plot_helper = PlotMatplot(env.observation_space)


            # plot the layout (position of each elements) of the powergrid
            plot_helper.plot_layout()

            # project some data on the grid
            line_values = env.get_thermal_limit()
            plot_helper.plot_info(line_values=line_values)

            # to plot an observation
            obs = env.reset()
            plot_helper.plot_obs(obs)


        Returns
        -------
        res: ``object``
            The figure updated with the data from the new observation.
        """
        # Check values are in the correct format
        if (
            line_values is not None
            and len(line_values) != self.observation_space.n_line
        ):
            raise PlotError(
                "Impossible to display these values on the powerlines: there are {} values"
                "provided for {} powerlines in the grid".format(
                    len(line_values), self.observation_space.n_line
                )
            )
        if (
            load_values is not None
            and len(load_values) != self.observation_space.n_load
        ):
            raise PlotError(
                "Impossible to display these values on the loads: there are {} values"
                "provided for {} loads in the grid".format(
                    len(load_values), self.observation_space.n_load
                )
            )
        if gen_values is not None and len(gen_values) != self.observation_space.n_gen:
            raise PlotError(
                "Impossible to display these values on the generators: there are {} values"
                "provided for {} generators in the grid".format(
                    len(gen_values), self.observation_space.n_gen
                )
            )
        if (
            storage_values is not None
            and len(storage_values) != self.observation_space.n_storage
        ):
            raise PlotError(
                "Impossible to display these values on the storage units: there are {} values"
                "provided for {} generators in the grid".format(
                    len(storage_values), self.observation_space.n_storage
                )
            )

        # Get a valid figure to draw into
        if figure is None:
            fig = self.create_figure()
            redraw = True
        elif redraw:
            self.clear_figure(figure)
            fig = figure
        else:
            fig = figure

        # Get a valid Observation
        if observation is None:
            # See dummy data added in the constructor
            observation = self.observation_space
            if coloring is not None:
                observation = copy.deepcopy(observation)
                if coloring == "line":
                    if line_values is None:
                        raise PlotError(
                            "Impossible to color the grid based on the line information (key word argument "
                            '"line_values") if this argument is None.'
                        )

                    observation.rho = copy.deepcopy(line_values)
                    try:
                        observation.rho = np.array(observation.rho).astype(dt_float)
                    except:
                        raise PlotError(
                            "Impossible to convert the input values (line_values) to floating point"
                        )

                    # rescaling to have range 0 - 1.0
                    tmp = observation.rho[np.isfinite(observation.rho)]
                    observation.rho -= (
                        np.min(tmp) - 1e-1
                    )  # so the min is 1e-1 otherwise 0.0 is plotted as black
                    tmp = observation.rho[np.isfinite(observation.rho)]
                    observation.rho /= np.max(tmp)
                elif coloring == "load":
                    # TODO
                    warnings.warn("coloring = loads is not available at the moment")
                elif coloring == "gen":
                    if gen_values is None:
                        raise PlotError(
                            "Impossible to color the grid based on the gen information (key word argument "
                            '"gen_values") if this argument is None.'
                        )

                    observation.prod_p = copy.deepcopy(gen_values)
                    try:
                        observation.prod_p = np.array(observation.prod_p).astype(
                            dt_float
                        )
                    except:
                        raise PlotError(
                            "Impossible to convert the input values (gen_values) to floating point"
                        )

                    # rescaling to have range 0 - 1.0
                    tmp = observation.prod_p[np.isfinite(observation.prod_p)]
                    if (np.isfinite(observation.prod_p)).any():
                        observation.prod_p -= (
                            np.min(tmp) - 1e-1
                        )  # so the min is 1e-1 otherwise 0.0 is plotted as black
                        tmp = observation.prod_p[np.isfinite(observation.prod_p)]
                        observation.prod_p /= np.max(tmp)
                else:
                    raise PlotError('coloring must be one of "line", "load" or "gen"')

        # Trigger draw calls
        self._plot_lines(fig, observation, line_values, line_unit, redraw)
        self._plot_loads(fig, observation, load_values, load_unit, redraw)
        self._plot_storages(fig, observation, storage_values, storage_unit, redraw)
        self._plot_gens(fig, observation, gen_values, gen_unit, redraw)
        self._plot_subs(fig, observation, redraw)
        self._plot_legend(fig, observation, redraw)

        # Some implementations may need postprocessing
        self.plot_postprocess(fig, observation, not redraw)

        # Return updated figure
        return fig



# Copyright (c) 2021, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.
import copy
import numpy as np

from grid2op.dtypes import dt_float, dt_int, dt_bool
from grid2op.Exceptions import NotEnoughAttentionBudget


class LinearAttentionBudget:
    def __init__(self):
        self._max_budget = None
        self._budget_per_ts = None
        self._alarm_cost = None
        self._current_budget = None
        self._init_budget = None
        self._time_last_alarm_raised = dt_int(-1)
        self._time_last_successful_alarm_raised = dt_int(-1)
        self._last_alarm_raised = None
        self._last_successful_alarm_raised = None
        self._all_successful_alarms = []

    @property
    def time_last_alarm_raised(self):
        """
        Time step of the last alarm raised (-1 if no alarm has been raised yet)
        Returns
        -------

        """
        return self._time_last_alarm_raised

    @property
    def time_last_successful_alarm_raised(self):
        """time of the last successful alarm raised"""
        return self._time_last_successful_alarm_raised

    @property
    def current_budget(self):
        """current attention budget"""
        return self._current_budget

    @property
    def last_alarm_raised(self):
        """
        for each zone, says:

          - -1 if no alarm have been raised for this zone for the entire episode
          - `k` (with `k>0`) says that the last alarm raised for this zone was at step `k`

        .. note::
            This counts both successful and non successful alarms
        """
        return self._last_alarm_raised

    @property
    def last_successful_alarm_raised(self):
        """
        for each zone, says:

          - -1 if no alarm have been raised for this zone for the entire episode
          - `k` (with `k>0`) says that the last alarm raised for this zone was at step `k`

        .. note::
            This counts only successful alarms

        """
        return self._last_successful_alarm_raised

    def init(
        self, partial_env, init_budget, max_budget, budget_per_ts, alarm_cost, **kwargs
    ):
        self._max_budget = dt_float(max_budget)
        self._budget_per_ts = dt_float(budget_per_ts)
        self._alarm_cost = dt_float(alarm_cost)
        self._init_budget = dt_float(init_budget)
        self._last_alarm_raised = np.empty(partial_env.dim_alarms, dtype=dt_int)
        self._last_successful_alarm_raised = np.empty(
            partial_env.dim_alarms, dtype=dt_int
        )
        self.reset()

    def reset(self):
        """
        called each time the scenario is over by the environment

        Returns
        -------

        """
        self._current_budget = self._init_budget
        self._time_last_alarm_raised = dt_int(-1)
        self._time_last_successful_alarm_raised = dt_int(-1)
        self._last_alarm_raised[:] = -1
        self._last_successful_alarm_raised[:] = -1
        self._all_successful_alarms = []

    def get_state(self):
        """used to retrieve the sate in simulate"""
        res = (
            self._time_last_alarm_raised,
            self._last_alarm_raised,
            self._current_budget,
            self._time_last_successful_alarm_raised,
            self._last_successful_alarm_raised,
            self._all_successful_alarms,
        )
        return res

    def set_state(self, state):
        """used to update the internal state of the budget, for simulate"""
        (
            _time_last_alarm_raised,
            _last_alarm_raised,
            _current_budget,
            _time_last_successful_alarm_raised,
            _last_successful_alarm_raised,
            _all_successful_alarms,
        ) = state

        self._time_last_alarm_raised = _time_last_alarm_raised
        self._last_alarm_raised[:] = _last_alarm_raised
        self._current_budget = _current_budget
        self._time_last_successful_alarm_raised = _time_last_successful_alarm_raised
        self._last_successful_alarm_raised[:] = _last_successful_alarm_raised
        self._all_successful_alarms = copy.copy(_all_successful_alarms)

    def register_action(self, env, action, is_action_illegal, is_action_ambiguous):
        """
        INTERNAL

        Called at each step to update the budget according to the action played

        Parameters
        ----------
        env
        action
        is_action_illegal
        is_action_ambiguous

        Returns
        -------

        """
        if action.dim_alarms == 0 or is_action_illegal or is_action_ambiguous:
            # this feature is not supported (grid2op <= 1.6.0) or is not activated

            # also, if the action is illegal is ambiguous, it is replaced with do nothing, but i don't really
            # want to affect the budget on this case
            return None

        if action.alarm_raised().size:
            # an alarm has been raised
            self._time_last_alarm_raised = env.nb_time_step
            self._last_alarm_raised[action.raise_alarm] = env.nb_time_step
            if self._current_budget >= self._alarm_cost:
                # i could raise it
                self._current_budget -= self._alarm_cost
                self._time_last_successful_alarm_raised = env.nb_time_step
                self._last_successful_alarm_raised[
                    action.raise_alarm
                ] = env.nb_time_step
                self._all_successful_alarms.append(
                    (env.nb_time_step, copy.deepcopy(action.raise_alarm))
                )
            else:
                # not enough budget
                current_budget = self._current_budget
                # self._current_budget = 0
                return NotEnoughAttentionBudget(
                    f"You need a budget of {self._alarm_cost} to raise an alarm "
                    f"but you had only {current_budget}. Nothing is done."
                )
        else:
            # no alarm has been raised, budget increases
            self._current_budget = min(
                self._max_budget, self._budget_per_ts + self._current_budget
            )
        return None

class _LinearAttentionBudgetByLine:
    """Currently not implemented, DO NOT USE !"""
    def __init__(self):
        self._max_budget = None
        self._budget_per_ts = None
        self._alert_cost = None
        self._current_budget = None
        self._init_budget = None
        self._time_last_alert_raised = dt_int(-1)
        self._time_last_successful_alert_raised = dt_int(-1)
        self._is_last_alert_successful = None
        self._time_window = None
        self._alert_size = None
        self._last_alert_action_filtered_by_budget = None

    @property
    def time_last_alert_raised(self):
        """
        Time step of the last alert raised (-1 if no alert has been raised yet)
        Returns
        -------

        """
        return self._time_last_alert_raised

    @property
    def time_last_successful_alert_raised(self):
        """time of the last successful alert raised"""
        return self._time_last_successful_alert_raised

    @property
    def current_budget(self):
        """current attention budget"""
        return self._current_budget

    @property
    def is_last_alert_successful(self):
        """
        boolean value telling whether the alert is successfully raised (according to compliance with the budget)
        """
        return self._is_last_alert_successful

    @property
    def last_alert_action_filtered_by_budget(self): 
        """the actual alert action, after filtering to be compliant with the budget constraint"""
        return self._last_alert_action_filtered_by_budget

    @property
    def time_window(self): 
        """ """
        return self._time_window

    def init(
        self, partial_env, init_budget, max_budget, budget_per_ts, alert_cost, **kwargs
    ):
        self._max_budget = dt_float(max_budget)
        self._budget_per_ts = dt_float(budget_per_ts)
        self._alert_cost = dt_float(alert_cost)
        self._init_budget = dt_float(init_budget)
        self._time_window = partial_env.parameters.ALERT_TIME_WINDOW
        self._dim_alerts = 0
        if "dim_alerts" in kwargs.keys():
            self._dim_alerts = kwargs["dim_alerts"]
        self.reset()

    def reset(self):
        """
        called each time the scenario is over by the environment

        Returns
        -------

        """
        self._current_budget = self._init_budget
        self._time_last_alert_raised = dt_int(-1)
        self._time_last_successful_alert_raised = dt_int(-1)
        self._is_last_alert_successful = False
        self._last_alert_action_filtered_by_budget = np.full(self._dim_alerts, False, dtype=dt_bool)
        

    def get_state(self):
        """used to retrieve the sate in simulate"""
        res = (
            self._time_last_alert_raised,
            self._current_budget,
            self._time_last_successful_alert_raised,
            self._is_last_alert_successful, 
            self._last_alert_action_filtered_by_budget
        )
        return res

    def set_state(self, state):
        """used to update the internal state of the budget, for simulate"""
        (
            _time_last_alert_raised,
            _current_budget,
            _time_last_successful_alert_raised,
            _is_last_alert_successful,
            _last_alert_action_filtered_by_budget
        ) = state

        self._time_last_alert_raised = _time_last_alert_raised
        self._current_budget = _current_budget
        self._time_last_successful_alert_raised = _time_last_successful_alert_raised
        self._is_last_alert_successful = _is_last_alert_successful
        self._last_alert_action_filtered_by_budget = _last_alert_action_filtered_by_budget

    def register_action(self, env, action, is_action_illegal, is_action_ambiguous):
        """
        INTERNAL

        Called at each step to update the budget according to the action played

        Parameters
        ----------
        env
        action
        is_action_illegal
        is_action_ambiguous

        Returns
        -------

        """
        if action.dim_alerts == 0 or is_action_illegal or is_action_ambiguous:
            # this feature is not supported (grid2op <= 1.6.0) or is not activated

            # also, if the action is illegal is ambiguous, it is replaced with do nothing, but i don't really
            # want to affect the budget on this case

            self._is_last_alert_successful = False
            self._last_alert_action_filtered_by_budget[:] = False

            return None

        self._is_last_alert_successful = False
        self._last_alert_action_filtered_by_budget[:] = False

        if action.alert_raised().size:
            
            # an alert has been raised
            self._time_last_alert_raised = env.nb_time_step

            self._is_last_alert_successful = True
            self._last_alert_action_filtered_by_budget = action._raise_alert

            nb_of_alerts = sum(action._raise_alert)
            if self._current_budget >= self._alert_cost * nb_of_alerts :
                # The alert is raisable 
                self._current_budget -= self._alert_cost * nb_of_alerts
                self._time_last_successful_alert_raised = env.nb_time_step
                
            else:
                # not enough budget
                current_budget = self._current_budget
                # self._current_budget = 0
                return NotEnoughAttentionBudget(
                    f"You need a budget of {self._alert_cost} to raise an alert "
                    f"but you had only {current_budget}. Nothing is done."
                )
        else:
            # no alert has been raised, budget increases
            self._current_budget = min(
                self._max_budget, self._budget_per_ts + self._current_budget
            )
        return None



# Copyright (c) 2023, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

import numpy as np
from grid2op.Reward import AlertReward
from grid2op.dtypes import dt_float

class _AlertTrustScore(AlertReward):
    """

    INTERNAL

    .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\
            It **must not** serve as a reward. This scored needs to be **MAXIMIZED**,
            as it is a negative! Also, this "reward" is not scaled or anything. Use it as your
            own risk.

    Implemented as a reward to make it easier to use in the context of the L2RPN competitions, this reward is based on the "alert feature" 
    where the agent is asked to send information about potential line overload issue on the grid after unpredictable powerline
    disconnection (attack of the opponent).
    The alerts are assessed once per attack and cumulated over the episode. In this scheme, this "reward" computed the assistant "score",
    which assesses how well the agent is aware of its capacities to deal with a situation during an episode.
    It should not be used to train an agent.
    
    """

    def __init__(self,
                 logger=None,
                 reward_min_no_blackout=-1.0,
                 reward_min_blackout=-50,
                 reward_max_no_blackout=0.0,
                 reward_max_blackout=0.0,
                 reward_end_episode_bonus=0.0,
                 min_score=-3):

        super().__init__(logger,
                 reward_min_no_blackout,
                 reward_min_blackout, 
                 reward_max_no_blackout,
                 reward_max_blackout,
                 reward_end_episode_bonus)
        
        self.min_score = dt_float(min_score)
        self.max_score = dt_float(1.0)
        self.blackout_encountered = False
        
    def initialize(self, env):
        self._is_simul_env = self.is_simulated_env(env)
            
        self.cumulated_reward = 0
        #KPIs
        self.total_nb_attacks = 0
        self.nb_last_attacks = 0 #attacks in the AlertTimeWindow before done

        #TODO
        #self.total_nb_alerts = 0
        #self.alert_attack_no_blackout = 0
        #self.alert_attack_blackout = 0
        #self.no_alert_attack_no_blackout = 0
        #self.no_alert_attack_blackout = 0
        #self.blackout_encountered = False
        return super().initialize(env)
            
    def reset(self, env):
        super().reset(env)
        self.blackout_encountered = False
        self.cumulated_reward = 0
        #KPIs
        self.total_nb_attacks = 0
        self.nb_last_attacks = 0

        # TODO
        #self.total_nb_alerts = 0
        #self.alert_attack_no_blackout = 0
        #self.alert_attack_blackout = 0
        #self.no_alert_attack_no_blackout = 0
        #self.no_alert_attack_blackout = 0
        #self.blackout_encountered = False
            
    def __call__(self, action, env, has_error, is_done, is_illegal, is_ambiguous):

        score_ep = 0.
        if self._is_simul_env:
            return score_ep

        self.blackout_encountered = self.is_in_blackout(has_error, is_done)
            
        res = super().__call__(action, env, has_error, is_done, is_illegal, is_ambiguous)
        self.cumulated_reward += res

        is_attack = (env._time_since_last_attack == 0).any()#even if there are simultaneous attacks, we consider this as a single attack event
        self.total_nb_attacks += is_attack

        # TODO
        #lines_alerted_beforeattack = np.equal(env._time_since_last_alert, env._time_since_last_attack + 1) and lines_attacked
        #self.total_nb_alerts += np.sum(lines_alerted_beforeattack)
            
        if is_done:
            indexes_to_look = (np.arange(-self.time_window, 1) + self._current_id) % self._nrows_array  # include current step (hence the np.arange(..., **1**))
            ts_attack_in_order = self._ts_attack[indexes_to_look, :]
            self.nb_last_attacks = ts_attack_in_order.any(axis=1).sum()

            cm_reward_min_ep, cm_reward_max_ep = self._compute_min_max_reward(self.total_nb_attacks,
                                                                              self.nb_last_attacks)
            score_ep = self._normalisation_fun(self.cumulated_reward, cm_reward_min_ep, cm_reward_max_ep,
                                               self.min_score, self.max_score, self.blackout_encountered)

            # TODO
            # if self.blackout_encountered:
            #    lines_attacked_dangerzone = (env._time_since_last_attack >= 0) * (env._time_since_last_attack < SURVIVOR_TIMESTEPS)
            #
            #    self.alert_attack_blackout += 1. * any(lines_alerted_beforeattack[lines_attacked_dangerzone])
            #    self.no_alert_attack_blackout += 1. * any(~lines_alerted_beforeattack[lines_attacked_dangerzone])
            # else :
            #    lines_attacked_no_blackout = env._time_since_last_attack > 0
            #
            #    self.alert_attack_no_blackout += np.sum(lines_alerted_beforeattack[lines_attacked_no_blackout])
            #    self.no_alert_attack_no_blackout += np.sum(~lines_alerted_beforeattack[lines_attacked_no_blackout])

            return score_ep
            
        else:
            # TODO
            # lines_attacked_no_blackout = env._time_since_last_attack == SURVIVOR_TIMESTEPS
            #
            # self.alert_attack_no_blackout += np.sum(lines_alerted_beforeattack[lines_attacked_no_blackout])
            # self.no_alert_attack_no_blackout += np.sum(~lines_alerted_beforeattack[lines_attacked_no_blackout])

            return score_ep
        
    @staticmethod
    def _normalisation_fun(cm_reward, cm_reward_min_ep, cm_reward_max_ep,min_score,max_score,is_blackout):

        #in case cm_reward_min_ep=cm_reward_max_ep=0, score can be maximum or 0
        if cm_reward_min_ep == cm_reward_max_ep:
            if is_blackout:
                score_ep = 0.0
            else:
                score_ep = max_score
        else:
            standardized_score = (cm_reward - cm_reward_min_ep) / (cm_reward_max_ep - cm_reward_min_ep)#denominator cannot be 0 given first if condition
            score_ep = min_score + (max_score - min_score) * standardized_score
        return score_ep
    
    def _compute_min_max_reward(self, nb_attacks,nb_last_attacks):

        if self.blackout_encountered:
            if nb_last_attacks == 0: #in the case the blackout is not tied to a recent attack
                cm_reward_min_ep = self.reward_min_no_blackout * nb_attacks
                cm_reward_max_ep = self.reward_max_no_blackout * nb_attacks
            elif nb_last_attacks >= 1:#in the case the blackout can be tied to one or several recent attacks 
                cm_reward_min_ep = self.reward_min_no_blackout * (nb_attacks - nb_last_attacks) + self.reward_min_blackout
                cm_reward_max_ep = self.reward_max_no_blackout * (nb_attacks - nb_last_attacks) + self.reward_max_blackout
        else:
            cm_reward_min_ep = self.reward_min_no_blackout * nb_attacks
            cm_reward_max_ep = self.reward_max_no_blackout * nb_attacks + self.reward_end_episode_bonus
        
        return cm_reward_min_ep, cm_reward_max_ep
        



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

import numpy as np
import copy

from grid2op.Exceptions import Grid2OpException
from grid2op.Reward import AlarmReward
from grid2op.dtypes import dt_float


class _AlarmScore(AlarmReward):
    """
        .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\
            It **must not** serve as a reward. This scored needs to be minimized,
            and a reward needs to be maximized! Also, this "reward" is not scaled or anything. Use it as your
            own risk.

    Implemented as a reward to make it easier to use in the context of the L2RPN competitions, this "reward"
    computed the "grid operation cost". It should not be used to train an agent.

    The "reward" the closest to this score is given by the :class:`AlarmReward` class.

    This reward is based on the "alarm feature" where the agent is asked to send information about potential issue
    on the grid.
    On this case, when the environment is in a "game over" state (eg it's the end) then the reward is computed
    the following way:
    - if the environment has been successfully manage until the end of the chronics, then 1.0 is returned
    - if no alarm has been raised, then -2.0 is return
    - points for pointing to the right zones are computed based on the lines disconnected either in a short window
    before game over or otherwise at the time of game over

    Examples
    ---------
    You can use this reward in any environment with:

    .. code-block:

        import grid2op
        from grid2op.Reward import AlarmReward
        # then you create your environment with it:
        NAME_OF_THE_ENVIRONMENT = "rte_case14_realistic"
        env = grid2op.make(NAME_OF_THE_ENVIRONMENT,reward_class=AlarmScore)
        # and do a step with a "do nothing" action
        obs = env.reset()
        obs, reward, done, info = env.step(env.action_space())
        # the reward is computed with the AlarmScore class

    """

    def __init__(self, logger=None):
        AlarmReward.__init__(self, logger=logger)
        # required if you want to design a custom reward taking into account the
        # alarm feature
        self.reward_min = dt_float(-2.0)
        # we keep other parameters values from AlarmReward as is

        self.mult_for_right_zone = 1.5
        self.window_disconnection = 4

        self.disc_lines_all_before_cascade = []
        self.n_line = None
        self._i_am_simulate = True

    def initialize(self, env):
        if not env._has_attention_budget:
            raise Grid2OpException(
                'Impossible to use the "_AlarmScore" with an environment for which this feature '
                'is disabled. Please make sure "env._has_attention_budget" is set to ``True`` or '
                "change the reward class with `grid2op.make(..., reward_class=AnyOtherReward)`"
            )
        self.n_line = env.n_line
        self.reset(env)

    def reset(self, env):
        super().reset(env)
        self.window_disconnection = max(self.best_time - self.window_size, 4)
        self.disc_lines_all_before_cascade = []
        self._i_am_simulate = self.is_simulated_env(env)

    def _lines_disconnected_first(self, disc_lines_at_cascading_time):
        """
        here we detect the disconnected lines that we will consider to compute the `mult_for_zone` multiplying factor.
        Either the lines that were disconnected in a short period before final failure. Otherwise the first lines
        disconnected at the time of failure

        :param disc_lines_at_cascading_time: lines that are disconnected first at time of failure
        :return:
        """

        disc_lines_to_consider_for_score = np.zeros(self.n_line, dtype=bool)

        nb_obs = len(self.disc_lines_all_before_cascade)
        for step in range(nb_obs - self.window_disconnection, nb_obs):
            disc_lines_to_consider_for_score[
                self.disc_lines_all_before_cascade[step] >= 0
            ] = True

        if disc_lines_to_consider_for_score.sum() == 0:
            disc_lines_to_consider_for_score = disc_lines_at_cascading_time == 0

        # if we are there, it is because we have identified before that the failure is due to disconnected powerlines
        assert (disc_lines_to_consider_for_score).any()

        # we transform the vector so that disconnected lines have a zero, to be coherent with env._disc_lines
        return 1 - disc_lines_to_consider_for_score

    def __call__(self, action, env, has_error, is_done, is_illegal, is_ambiguous):
        if self._i_am_simulate:
            return self.reward_no_game_over

        disc_lines_now = env._disc_lines

        if is_done:
            if not has_error:
                # agent went until the end
                return self.reward_max
            if np.all(env._disc_lines == -1):
                # game over is not caused by the tripping of a powerline
                return self.reward_min

            if len(env._attention_budget._all_successful_alarms) == 0:
                # no alarm have been sent, so it's the minimum
                return self.reward_min

            successfull_alarms = env._attention_budget._all_successful_alarms
            step_game_over = env.nb_time_step

            disc_lines_to_consider_for_score = self._lines_disconnected_first(
                disc_lines_now
            )

            # so now i can consider the alarms.
            best_score, is_alarm_used = self.reward_min, False
            for alarm in successfull_alarms:
                tmp_sc, tmp_is = self._points_for_alarm(
                    *alarm,
                    step_game_over=step_game_over,
                    disc_lines=disc_lines_to_consider_for_score,
                    env=env
                )
                if tmp_sc > best_score:
                    best_score = tmp_sc
                    is_alarm_used = tmp_is

            self.is_alarm_used = is_alarm_used
            return best_score
        else:
            # make sure to deepcopy, otherwise it gets updated with the last timestep value for every previous timesteps
            # we log the line disconnected over time
            # TODO have a cache there and store only the last few states, most of what is stored here is not used
            self.disc_lines_all_before_cascade.append(copy.deepcopy(disc_lines_now))
            res = self.reward_no_game_over
        return res



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

import numpy as np

from grid2op.dtypes import dt_int, dt_float
from grid2op.Observation.baseObservation import BaseObservation
from grid2op.Observation.completeObservation import CompleteObservation


class NoisyObservation(BaseObservation):
    """
    This class represent a complete observation (in the sens that all attributes
    of an :attr:`CompleteObservation` are accessible) but some of them are
    "noisy".

    That is, the observation that the agent has access to is not
    exactly the same as the environment internal values.

    The affected attributes are :

    - load_p: \*= lognormal (to keep the sign)
    - load_q: \*= lognormal (to keep the sign)
    - gen_p: \*= lognormal (to keep the sign)
    - gen_q: \*= lognormal (to keep the sign)
    - p_or += normal
    - p_ex += normal
    - q_or += normal
    - q_ex += normal
    - a_or: \*= lognormal (to keep the sign)
    - a_ex: \*= lognormal (to keep the sign)
    - rho: same noise as a_or (because rho is not "physical" it's the result of a computation)
    - storage_power += normal

    It can be used to emuate the acquisition of data coming from noisy sensors for
    example.

    Examples
    --------

    It can be used as follow:

    .. code-block:: python

        import grid2op

        env_name = "l2rpn_case14_sandbox"  # or any other name
        kwargs_observation = {"sigma_load_p": 0.1, "sigma_gen_p": 1.0}  # noise of the observation
        env = grid2op.make(env_name,
                           observation_class=NoisyObservation,
                           kwargs_observation=kwargs_observation)

        # do whatever you want with env !

    """

    attr_list_vect = CompleteObservation.attr_list_vect
    attr_list_json = CompleteObservation.attr_list_json
    attr_list_set = CompleteObservation.attr_list_set

    def __init__(
        self,
        obs_env=None,
        action_helper=None,
        random_prng=None,
        kwargs_env=None,
        sigma_load_p=0.01,  # multiplicative (log normal)
        sigma_load_q=0.01,  # multiplicative (log normal)
        sigma_gen_p=0.01,  # multiplicative (log normal)
        sigma_gen_q=0.01,  # multiplicative (log normal)
        sigma_a=0.01,  # multiplicative (log normal) same for a_or and a_ex
        sigma_p=0.1,  # additive (normal) same for p_or and p_ex
        sigma_q=0.1,  # additive (normal) same for q_or and q_ex
        sigma_storage=0.1,  # additive (normal)
    ):

        BaseObservation.__init__(
            self,
            obs_env=obs_env,
            action_helper=action_helper,
            random_prng=random_prng,
            kwargs_env=kwargs_env
        )
        self._dictionnarized = None
        self._sigma_load_p = sigma_load_p  # multiplicative (log normal)
        self._sigma_load_q = sigma_load_q  # multiplicative (log normal)
        self._sigma_gen_p = sigma_gen_p  # multiplicative (log normal)
        self._sigma_gen_q = sigma_gen_q  # multiplicative (log normal)
        self._sigma_a = sigma_a  # multiplicative (log normal) same for a_or and a_ex

        self._sigma_p = sigma_p  # additive (normal) same for p_or and p_ex
        self._sigma_q = sigma_q  # additive (normal) same for q_or and q_ex
        self._sigma_storage = sigma_storage  # additive (normal)

    def update(self, env, with_forecast=True):
        # reset the matrices
        self._reset_matrices()
        self.reset()

        # update as if the data were complete
        self._update_obs_complete(env, with_forecast=with_forecast)

        # multiplicative noise
        mult_load_p = self.random_prng.lognormal(
            mean=0.0, sigma=self._sigma_load_p, size=self.load_p.shape
        )
        self.load_p[:] *= mult_load_p
        mult_load_q = self.random_prng.lognormal(
            mean=0.0, sigma=self._sigma_load_q, size=self.load_p.shape
        )
        self.load_q[:] *= mult_load_q
        mult_gen_p = self.random_prng.lognormal(
            mean=0.0, sigma=self._sigma_gen_p, size=self.gen_p.shape
        )
        self.gen_p[:] *= mult_gen_p
        mult_gen_q = self.random_prng.lognormal(
            mean=0.0, sigma=self._sigma_gen_q, size=self.gen_q.shape
        )
        self.gen_q[:] *= mult_gen_q
        mult_aor = self.random_prng.lognormal(
            mean=0.0, sigma=self._sigma_a, size=self.a_or.shape
        )
        self.a_or[:] *= mult_aor
        self.rho[:] *= mult_aor
        mult_a_ex = self.random_prng.lognormal(
            mean=0.0, sigma=self._sigma_a, size=self.a_ex.shape
        )
        self.a_ex[:] *= mult_a_ex

        # additive noise
        add_por = self.random_prng.normal(
            loc=0.0,
            scale=self._sigma_p,  # 0.01 * np.abs(self.p_or),
            size=self.p_or.shape,
        )
        self.p_or[:] += add_por

        add_pex = self.random_prng.normal(
            loc=0.0,
            scale=self._sigma_p,  # 0.01 * np.abs(self.p_ex),
            size=self.p_or.shape,
        )
        self.p_ex[:] += add_pex

        add_qor = self.random_prng.normal(
            loc=0.0,
            scale=self._sigma_q,  # 0.01 * np.abs(self.q_or),
            size=self.p_or.shape,
        )
        self.q_or[:] += add_qor

        add_qex = self.random_prng.normal(
            loc=0.0,
            scale=self._sigma_q,  # 0.01 * np.abs(self.q_ex),
            size=self.p_or.shape,
        )
        self.q_ex[:] += add_qex

        add_storp = self.random_prng.normal(
            loc=0.0,
            scale=self._sigma_storage,  # 0.01 * np.abs(self.storage_power),
            size=self.storage_power.shape,
        )
        self.storage_power[:] += add_storp



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.
import numpy as np
import math

from grid2op.Converter import Converter


class AnalogStateConverter(Converter):
    """
    Converter that can be used with analog representation of the grid state.
    Details are provided in convert_obs and convert_act

    The grid2op observation is converted into a 1d normalied array
    The grid2op action is created from a set of real valued arrays

    It can not yet be converted to / from gym space. If this feature is interesting for you, you can
    reply to the issue posted at https://github.com/rte-france/Grid2Op/issues/16

    """

    def __init__(self, action_space, bias=0.0):
        super().__init__(action_space)
        self.__class__ = AnalogStateConverter.init_grid(action_space)
        self.__bias = 0.0

    @staticmethod
    def to_norm_vect(inputv, pad_v=0.0, scale_v=1.0):
        v = np.asarray(inputv)
        v = v / scale_v
        vsafe = np.nan_to_num(v, nan=pad_v, posinf=pad_v, neginf=pad_v)
        return vsafe.astype(np.float32)

    def convert_obs(self, obs):
        """
        This converter will convert the observation into a 1D vector,
        with all values normalized, plus bias (if provided)

        Parameters
        ----------
        obs: :class:`grid2op.Observation.Observation`
            The input observation.

        Returns
        -------
        ``np.array`` 1D array of np.float32 normalized values

        """
        # Store the obs for action conversion
        self.__obs = obs
        # Store some shortcuts
        topo = obs.topo_vect
        g_pos = obs.gen_pos_topo_vect
        l_pos = obs.load_pos_topo_vect
        lor_pos = obs.line_or_pos_topo_vect
        lex_pos = obs.line_ex_pos_topo_vect

        # Get time data
        time_li = [
            obs.month / 12.0,
            obs.day / 31.0,
            obs.day_of_week / 7.0,
            obs.hour_of_day / 24.0,
            obs.minute_of_hour / 60.0,
        ]
        time_v = self.to_norm_vect(time_li)
        time_line_cd = self.to_norm_vect(
            obs.time_before_cooldown_line, pad_v=-1.0, scale_v=10.0
        )
        time_line_nm = self.to_norm_vect(obs.time_next_maintenance, scale_v=10.0)
        time_sub_cd = self.to_norm_vect(
            obs.time_before_cooldown_sub, pad_v=-1.0, scale_v=10.0
        )

        # Get generators info
        g_p = self.to_norm_vect(obs.prod_p, scale_v=1000.0)
        g_q = self.to_norm_vect(obs.prod_q, scale_v=1000.0)
        g_v = self.to_norm_vect(obs.prod_v, scale_v=1000.0)
        g_tr = self.to_norm_vect(obs.target_dispatch, scale_v=150.0)
        g_ar = self.to_norm_vect(obs.actual_dispatch, scale_v=150.0)
        g_cost = self.to_norm_vect(obs.gen_cost_per_MW, pad_v=0.0, scale_v=1.0)
        g_buses = np.zeros(obs.n_gen)
        for gen_id in range(obs.n_gen):
            g_buses[gen_id] = topo[g_pos[gen_id]]
            if g_buses[gen_id] <= 0.0:
                g_buses[gen_id] = 0.0
        g_bus = self.to_norm_vect(g_buses, pad_v=-1.0, scale_v=3.0)

        # Get loads info
        l_p = self.to_norm_vect(obs.load_p, scale_v=1000.0)
        l_q = self.to_norm_vect(obs.load_q, scale_v=1000.0)
        l_v = self.to_norm_vect(obs.load_v, scale_v=1000.0)
        l_buses = np.zeros(obs.n_load)
        for load_id in range(obs.n_load):
            l_buses[load_id] = topo[l_pos[load_id]]
            if l_buses[load_id] <= 0.0:
                l_buses[load_id] = 0.0
        l_bus = self.to_norm_vect(l_buses, pad_v=-1.0, scale_v=3.0)

        # Get lines origin info
        or_p = self.to_norm_vect(obs.p_or, scale_v=1000.0)
        or_q = self.to_norm_vect(obs.q_or, scale_v=1000.0)
        or_v = self.to_norm_vect(obs.v_or, scale_v=1000.0)
        or_buses = np.zeros(obs.n_line)
        for line_id in range(obs.n_line):
            or_buses[line_id] = topo[lor_pos[line_id]]
            if or_buses[line_id] <= 0.0:
                or_buses[line_id] = 0.0
        or_bus = self.to_norm_vect(or_buses, pad_v=-1.0, scale_v=3.0)
        or_rho = self.to_norm_vect(obs.rho, pad_v=-1.0)

        # Get extremities origin info
        ex_p = self.to_norm_vect(obs.p_ex, scale_v=1000.0)
        ex_q = self.to_norm_vect(obs.q_ex, scale_v=1000.0)
        ex_v = self.to_norm_vect(obs.v_ex, scale_v=1000.0)
        ex_buses = np.zeros(obs.n_line)
        for line_id in range(obs.n_line):
            ex_buses[line_id] = topo[lex_pos[line_id]]
            if ex_buses[line_id] <= 0.0:
                ex_buses[line_id] = 0.0
        ex_bus = self.to_norm_vect(ex_buses, pad_v=-1.0, scale_v=3.0)
        ex_rho = self.to_norm_vect(obs.rho, pad_v=-1.0)

        res = np.concatenate(
            [
                # Time
                time_v,
                time_line_cd,
                time_sub_cd,
                time_line_nm,
                # Gens
                g_p,
                g_q,
                g_v,
                g_ar,
                g_tr,
                g_bus,
                g_cost,
                # Loads
                l_p,
                l_q,
                l_v,
                l_bus,
                # Origins
                or_p,
                or_q,
                or_v,
                or_bus,
                or_rho,
                # Extremities
                ex_p,
                ex_q,
                ex_v,
                ex_bus,
                ex_rho,
            ]
        )
        return res + self.__bias

    def convert_act(self, netstate):
        """
        Create a grid2op action based on the last observation and
        the real valued state vectors in parameters

        Parameters
        ----------
        netstate: ``tuple``
            A tuple containing the following (3) elements:

            - netbus: ``np.array``
              A numpy array of dimension n_bus(2) x dim_topo and range [0.0; 1.0].

              Where the first axis represent the bus, the second the elements.
              Then, for element i, netbus[bus_index][i] represent the probability
              element i should be on bus_index + 1.
              The buses are then picked using argmax across dimension 0

            - netline: ``np.array``
              A numpy array of dimension n_line and range [0.0; 1.0]
              Each element representing a line status:
              0 meaning disconnected and > 0.0 connected

            - netdisp: ``np.array``
              A numpy array of dimension n_gen and range[-1.0;1.0]
              Each generator redispatch setpoint is then
              rescaled to the range [-rdown;+rup].
              This is cumulative over time, as per grid2op convention.

        Returns
        -------
        res: :class:`grid2op.Action.Action`
            An action that will change the last observation (current state)
            To the state described in parameters
        """
        netbus = netstate[0]
        netline = netstate[1]
        netdisp = netstate[2]
        act_setbus = self.netbus_to_act_setbus(self.__obs, netbus)
        act_setstatus = self.netline_to_act_setstatus(self.__obs, netline)
        act_redispatch = self.netdisp_to_act_redispatch(self.__obs, netdisp)
        act = self.__call__(
            {
                "set_bus": act_setbus,
                "set_line_status": act_setstatus,
                "redispatch": act_redispatch,
            }
        )
        return act

    @staticmethod
    def size_obs(obs):
        dims = np.array(
            [
                # Time
                5,  # Timestamp
                2 * obs.n_line,
                obs.n_sub,
                # Gen
                obs.n_gen * 7,
                # Load
                obs.n_load * 4,
                # Line origins
                obs.n_line * 5,
                # Line extremities
                obs.n_line * 5,
            ]
        )
        return dims.sum()

    @staticmethod
    def netbus_to_act_setbus(obs, net_bus):
        # n_bus x dim_topo x p([0.0; 1.0]) ->
        # -> dim_topo x [0 unchanged; 1: bus_1; 2 bus_2 ]
        # Pick the buses
        act_setbus = np.argmax(net_bus, axis=0) + 1
        # Don't set disconnected elements
        act_setbus[obs.topo_vect <= 0] = 0
        # Don't set elements already on the correct bus
        act_setbus[act_setbus == obs.topo_vect] = 0

        return act_setbus

    @staticmethod
    def netline_to_act_setstatus(obs, net_line):
        # [0.0 Disconnect; > 0.0 Connect] ->
        # -> [0.0 Unchanged; -1.0 Disconnect; 1.0 Connect]
        act_setstatus = np.copy(net_line)
        act_setstatus[net_line <= 0.0] = -1
        act_setstatus[net_line > 0.0] = 1
        # Do no 'set' already connected lines
        act_setstatus[obs.line_status == (act_setstatus == 1)] = 0
        # Do not 'set' already disconnected lines
        act_setstatus[(obs.line_status == False) == (act_setstatus == -1)] = 0
        return act_setstatus

    @staticmethod
    def netdisp_to_act_redispatch(obs, net_disp):
        # [-1.0;1.0] -> [-ramp_down;+ramp_up]
        act_redispatch = np.zeros(obs.n_gen)
        for i, d in enumerate(net_disp):
            if math.isclose(d, 0.0):  # Skip if 0.0
                continue
            rmin = obs.gen_max_ramp_down[i]
            rmax = obs.gen_max_ramp_up[i]
            r = np.interp(d, [-1.0, 1.0], [-rmin, rmax])
            act_redispatch[i] = round(r)  # Round at 1MW
        return act_redispatch

    # Helpers to generate random actions
    @staticmethod
    def netbus_rnd(obs, n_bus=2):
        # Copy obs state
        rnd_topo = np.zeros((n_bus, obs.dim_topo))
        rnd_topo[0][obs.topo_vect == 1] = 1.0
        rnd_topo[1][obs.topo_vect == 2] = 1.0
        # Pick a random substation
        rnd_sub = np.random.randint(obs.n_sub)
        n_elem = obs.sub_info[rnd_sub]
        # Pick a random number of elements to change
        rnd_n_changes = np.random.randint(n_elem + 1)
        # Pick the elements to change at random
        rnd_sub_elems = np.random.randint(0, n_elem, rnd_n_changes)
        # Set the topo vect
        sub_topo_pos = obs.sub_info[0:rnd_sub].sum()
        for elem_pos in rnd_sub_elems:
            rnd_bus = np.random.randint(n_bus)
            rnd_topo[rnd_bus][sub_topo_pos + elem_pos] = 1.0
            # Set the other buses to 0.0
            for b in range(n_bus):
                if b == rnd_bus:
                    continue
                rnd_topo[b][sub_topo_pos + elem_pos] = 0.0

        return rnd_topo

    @staticmethod
    def netline_rnd(obs):
        rnd_lines = obs.line_status.astype(np.float32)
        rnd_lineid = np.random.randint(obs.n_line)
        rnd_linestatus = not obs.line_status[rnd_lineid]
        rnd_lines[rnd_lineid] = np.int32(rnd_linestatus)

        return rnd_lines

    @staticmethod
    def netdisp_rnd(obs):
        disp_rnd = np.zeros(obs.n_gen)
        # Take random gen to disp
        rnd_gen = np.random.randint(obs.n_gen)
        # Take a random disp
        rnd_ramp = np.random.uniform(-1.0, 1.0)
        disp_rnd[rnd_gen] = rnd_ramp

        return disp_rnd



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

"""
This module defines the :class:`Renderer` that is able to display the state (:class:`grid2op.BaseObservation.BaseObservation`)
of the powergrid on a dedicated window.

It is also able to output a 3d representation of this representation to be further used by other libraries to
output gifs for example.

"""

import numpy as np
import cmath
import math
import os
import time

from grid2op.Plot.BasePlot import BasePlot
from grid2op.Exceptions.PlotExceptions import PyGameQuit, PlotError

try:
    os.environ["PYGAME_HIDE_SUPPORT_PROMPT"] = "hide"
    import pygame

    can_plot = True
except Exception as e:
    can_plot = False
    pass


class Point:
    # https://codereview.stackexchange.com/questions/70143/drawing-a-dashed-line-with-pygame
    # constructed using a normal tupple
    def __init__(self, point_t=(0, 0)):
        self.x = float(point_t[0])
        self.y = float(point_t[1])

    # define all useful operators
    def __add__(self, other):
        return Point((self.x + other.x, self.y + other.y))

    def __sub__(self, other):
        return Point((self.x - other.x, self.y - other.y))

    def __mul__(self, scalar):
        return Point((self.x * scalar, self.y * scalar))

    def __div__(self, scalar):
        return Point((self.x / scalar, self.y / scalar))

    def __floordiv__(self, scalar):
        return Point((self.x / scalar, self.y / scalar))

    def __truediv__(self, scalar):
        return Point((self.x / scalar, self.y / scalar))

    def __len__(self):
        return int(math.sqrt(self.x**2 + self.y**2))

    # get back values in original tuple format
    def get(self):
        return (self.x, self.y)

    def to_cplx(self):
        return self.x + 1j * self.y

    @staticmethod
    def from_cplx(cplx):
        return Point((cplx.real, cplx.imag))


def _draw_dashed_line(surf, color, start_pos, end_pos, width=1, dash_length=10):
    # https://codereview.stackexchange.com/questions/70143/drawing-a-dashed-line-with-pygame
    origin = Point(start_pos)
    target = Point(end_pos)
    displacement = target - origin
    length = len(displacement)
    slope = displacement / length

    for index in range(0, int(length / dash_length), 2):
        start = origin + (slope * index * dash_length)
        end = origin + (slope * (index + 1) * dash_length)
        pygame.draw.line(surf, color, start.get(), end.get(), width)


def _draw_arrow(
    surf,
    color,
    start_pos,
    end_pos,
    positive_flow,
    width=1,
    num_arrows=10,
    length_arrow=10,
    angle_arrow=30,
):
    if positive_flow:
        origin = Point(start_pos)
        target = Point(end_pos)
    else:
        target = Point(start_pos)
        origin = Point(end_pos)

    displacement = target - origin
    length = len(displacement)
    slope = displacement / length

    # phi = cmath.phase(slope.to_cplx()) * 360 / 2*cmath.pi
    phi = cmath.phase(displacement.to_cplx()) * 360 / (2 * cmath.pi)
    cste_ = 2 * cmath.pi / 360 * 1j
    rotatedown = cmath.exp(cste_ * (180 + phi + angle_arrow))
    rotateup = cmath.exp(cste_ * (180 + phi - angle_arrow))

    first_arrow_part = length_arrow * rotateup
    second_arrow_part = length_arrow * rotatedown

    per_displ = displacement / (num_arrows + 1)

    for index in range(0, int(num_arrows)):
        mid = origin + (per_displ * (index + 1))
        start_arrow = Point.from_cplx(mid.to_cplx() + first_arrow_part)
        end_arrow = Point.from_cplx(mid.to_cplx() + second_arrow_part)
        # , end_arrow.get()
        pygame.draw.lines(
            surf, color, False, [start_arrow.get(), mid.get(), end_arrow.get()], width
        )


class PlotPyGame(BasePlot):
    """
    INTERNAL

    .. warning:: /!\\\\ This class is deprecated /!\\\\

    This renderer should be used only for "online" representation of a powergrid.

    """

    def __init__(
        self,
        observation_space,
        substation_layout=None,
        radius_sub=20.0,
        load_prod_dist=70.0,
        bus_radius=5.0,
        timestep_duration_seconds=1.0,
        fontsize=20,
    ):
        """

        Parameters
        ----------
        substation_layout: ``list``
            List of tupe given the position of each of the substation of the powergrid.

        observation_space: :class:`grid2op.Observation.ObservationSpace`
            BaseObservation space used for the display

        radius_sub: ``int``
            radius (in pixel) of the substations representation.

        load_prod_dist: ``int``
            distance (in pixels) between the substation and the load or the generator.

        bus_radius: ``int``
            The buses are represented by small circles. This is the radius (in pixel) for the pixels representing
            the buses.

        timestep_duration_seconds: ``float``
            Minimum time during which a time step will stay on the screen, in second.

        fontsize: ``int``
            size of the font used to display the texts.


        """
        if not can_plot:
            raise PlotError("Impossible to plot as pygame cannot be imported.")

        self.window_grid = (1000, 700)
        self.lag_x = 150
        self.lag_y = 100

        BasePlot.__init__(
            self,
            substation_layout=substation_layout,
            observation_space=observation_space,
            radius_sub=radius_sub,
            load_prod_dist=load_prod_dist,
            bus_radius=bus_radius,
        )

        # pygame
        self.__is_init = False
        self.video_width, self.video_height = 1300, 700
        self.timestep_duration_seconds = timestep_duration_seconds
        self.time_last = None
        self.fontsize = fontsize
        self.background_color = [70, 70, 73]

        # init pygame
        self.display_called = None
        self.screen = None
        self.font = None
        self.init_pygame()

        # pause button
        self.font_pause = pygame.font.Font(None, 30)
        self.color_text = pygame.Color(255, 255, 255)
        self.text_paused = self.font_pause.render("Game Paused", True, self.color_text)

        # maximum overflow possible
        self.rho_max = 2.0

        # utilities
        self.cum_reward = 0.0
        self.nb_timestep = 0

        # colors
        self.col_line = pygame.Color(0, 0, 255)
        self.col_sub = pygame.Color(255, 0, 0)
        self.col_load = pygame.Color(0, 0, 0)
        self.col_gen = pygame.Color(0, 255, 0)
        self.default_color = pygame.Color(0, 0, 0)

        # deactivate the display on the screen
        self._deactivate_display = False

    def change_duration_timestep_display(self, new_timestep_duration_seconds):
        """
        INTERNAL

         .. warning:: /!\\\\ This class is deprecated /!\\\\

        Change the duration on which the screen is displayed.
        """
        self.timestep_duration_seconds = new_timestep_duration_seconds

    def init_pygame(self):
        if self.__is_init is False:
            pygame.init()
            self.display_called = False
            self.screen = pygame.display.set_mode(
                (self.video_width, self.video_height), pygame.RESIZABLE
            )
            self.font = pygame.font.Font(None, self.fontsize)
            self.__is_init = True

    def reset(self, env):
        """
        INTERNAL

         .. warning:: /!\\\\ This class is deprecated /!\\\\

        Parameters
        ----------
        env: :class:`grid2op.Environment.Environment`
            The used environment.

        Returns
        -------

        """
        self.cum_reward = 0.0
        self.nb_timestep = 0
        self.rho_max = env.parameters.HARD_OVERFLOW_THRESHOLD

    def _get_sub_layout(self, init_layout):
        tmp = [(el1, -el2) for el1, el2 in init_layout]

        # then scale the grid to be on the window, with the proper margin (careful, margin are applied both left and r
        # and right, so count twice
        tmp_arr = np.array(tmp)
        min_ = tmp_arr.min(axis=0)
        max_ = tmp_arr.max(axis=0)
        b = min_
        a = max_ - min_
        res = [
            (
                int((el1 - b[0]) / a[0] * (self.window_grid[0] - 2 * self.lag_x))
                + self.lag_x,
                int((el2 - b[1]) / a[1] * (self.window_grid[1] - 2 * self.lag_y))
                + self.lag_y,
            )
            for el1, el2 in tmp
        ]
        return res

    def _event_looper(self, force=False):
        has_quit = False
        if self._deactivate_display:
            return force, has_quit
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                has_quit = True
                return force, has_quit
                # pygame.quit()
                # exit()
            elif event.type == pygame.KEYDOWN:
                if event.key == pygame.K_ESCAPE:
                    has_quit = True
                    return force, has_quit
                if event.key == pygame.K_SPACE:
                    self._get_plot_pause()
                    # pause_surface = self.draw_plot_pause()
                    # self.screen.blit(pause_surface, (320 + self.left_menu_shape[0], 320))
                    return not force, has_quit
        return force, has_quit

    def _press_key_to_quit(self):
        """
        INTERNAL

         .. warning:: /!\\\\ This class is deprecated /!\\\\

        This utility function waits for the player to press a key to exit the renderer (called when the episode is done)

        Returns
        -------
        res: ``bool``, ``bool``
            ``True`` if the human player closed the window, in this case it will stop the computation: no other episode
            will be computed. ``False`` otherwise.

        """
        if self._deactivate_display:
            return
        has_quit = False
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                has_quit = True
                return True, has_quit
            elif event.type == pygame.KEYDOWN:
                if event.key == pygame.K_ESCAPE:
                    has_quit = True
                    return True, has_quit
                if event.key == pygame.K_SPACE:
                    return True, has_quit
        return False, has_quit

    def close(self):
        """
        This method is called when the renderer should be close.
        """
        self.display_called = False
        try:
            self._quit_and_close()
        except PyGameQuit:
            pass

    def _get_plot_pause(self):
        position = 300
        start_pause = position + self.text_paused.get_height()
        end_pause = start_pause + 50
        y_text_left = self.window_grid[0] + 100
        self.screen.blit(self.text_paused, (y_text_left, 300))

        pygame.draw.line(
            self.screen,
            self.color_text,
            (y_text_left + self.text_paused.get_width() // 2 - 10, start_pause),
            (y_text_left + self.text_paused.get_width() // 2 - 10, end_pause),
            10,
        )
        pygame.draw.line(
            self.screen,
            self.color_text,
            (y_text_left + self.text_paused.get_width() // 2 + 10, start_pause),
            (y_text_left + self.text_paused.get_width() // 2 + 10, end_pause),
            10,
        )
        pygame.display.flip()

    def _draw_final_information(self, reward, done, timestamp):
        if done is not None:
            if done:
                text_label = "GAME OVER, press any key to continue to next episode."
                text_graphic = self.font.render(text_label, True, self.color_text)
                self.screen.blit(text_graphic, (self.window_grid[0] + 100, 100))
                text_label = "Total cumulated reward: {:.1f}".format(self.cum_reward)
                text_graphic = self.font.render(text_label, True, self.color_text)
                self.screen.blit(text_graphic, (self.window_grid[0] + 100, 130))
                text_label = "Total number timesteps: {:.1f}".format(self.nb_timestep)
                text_graphic = self.font.render(text_label, True, self.color_text)
                self.screen.blit(text_graphic, (self.window_grid[0] + 100, 160))

    def _quit_and_close(self):
        # pygame.reset_vars()
        # pygame.gameLoop()
        pygame.display.quit()
        # pygame.quit()
        self.display_called = None
        self.screen = None
        self.font = None
        self.__is_init = False
        raise PyGameQuit()

    def deactivate_display(self):
        self._deactivate_display = True

    def get_rgb(self, obs, reward=None, done=None, timestamp=None):
        """
        INTERNAL

         .. warning:: /!\\\\ This class is deprecated /!\\\\

        Computes and returns the rgb 3d array from an observation, and potentially other informations.

        Parameters
        ----------
        obs: :class:`grid2op.Observation.Observation`
            The observation to converte into a 3d array

        reward: ``float``
            The current reward

        done: ``bool``
            Whether this is the last frame of the episode.

        timestamp: ``datetime.datetime``
            The curent datetime corresponding to the observation

        Returns
        -------
        res: ``numpy.ndarray``
            The 3d representation of the observation that can then be converted to a gif, or an image using appropriate
            softwares.

        """
        self.plot_obs(obs, reward, done, timestamp)
        return pygame.surfarray.array3d(self.screen)

    def init_fig(self, fig, reward, done, timestamp):
        self.init_pygame()

        if not self.display_called:
            self.display_called = True
            self.screen.fill(self.background_color)
            pygame.display.set_caption("Grid2Op Renderer")  # Window title

        force, has_quit = self._event_looper(force=False)
        while force:
            force, has_quit = self._event_looper(force=force)
            pygame.time.wait(250)  # it's in ms

        if has_quit:
            self._quit_and_close()

        if reward is not None:
            self.cum_reward += reward
            self.nb_timestep += 1

        # The game is not paused anymore (or never has been), I can render the next surface

        if self.time_last is not None and self._deactivate_display is False:
            tmp = time.perf_counter()  # in second
            if tmp - self.time_last < self.timestep_duration_seconds:
                nb_sec_wait = int(
                    1000 * (self.timestep_duration_seconds - (tmp - self.time_last))
                )
                pygame.time.wait(nb_sec_wait)  # it's in ms
            self.time_last = time.perf_counter()
        else:
            self.time_last = time.perf_counter()
        self.screen.fill(self.background_color)

        if done is not None:
            if not done:
                # draw the generic information on the right part
                self._draw_generic_info(reward, done, timestamp)
            else:
                # inform user that it's over
                self._draw_final_information(reward, done, timestamp)

    def _post_process_obs(
        self, fig, reward, done, timestamp, subs, lines, loads, gens, topos
    ):
        """
        INTERNAL

         .. warning:: /!\\\\ This class is deprecated /!\\\\

        In canse of plotply, fig is whether the player press "quit" or not

        Parameters
        ----------
        fig
        subs
        lines
        loads
        gens
        topos

        Returns
        -------

        """
        self._draw_final_information(reward, done, timestamp)
        pygame.display.flip()
        if self._deactivate_display is False:
            if done:
                key_pressed = False
                while not key_pressed:
                    key_pressed, has_quit = self._press_key_to_quit()
                    # TODO that with fps !!!
                    pygame.time.wait(250)  # it's in ms
                self._quit_and_close()

    def _draw_generic_info(self, reward=None, done=None, timestamp=None):
        if reward is not None:
            text_label = "Instantaneous reward: {:.1f}".format(reward)
            text_graphic = self.font.render(text_label, True, self.color_text)
            self.screen.blit(text_graphic, (self.window_grid[0] + 100, 100))
            text_label = "Cumulated reward: {:.1f}".format(self.cum_reward)
            text_graphic = self.font.render(text_label, True, self.color_text)
            self.screen.blit(text_graphic, (self.window_grid[0] + 100, 130))
            text_label = "Number timesteps: {:.1f}".format(self.nb_timestep)
            text_graphic = self.font.render(text_label, True, self.color_text)
            self.screen.blit(text_graphic, (self.window_grid[0] + 100, 160))

        if timestamp is not None:
            text_label = "Date : {:%Y-%m-%d %H:%M}".format(timestamp)
            text_graphic = self.font.render(text_label, True, self.color_text)
            self.screen.blit(text_graphic, (self.window_grid[0] + 100, 200))

    def _draw_subs_one_sub(self, fig, sub_id, center, this_col, text):
        pygame.draw.circle(
            self.screen,
            self.color_text,
            [int(el) for el in center],
            int(self.radius_sub),
            2,
        )

        text_graphic = self.font.render(text, True, this_col)
        self._aligned_text(center, text_graphic, center)

    def _get_default_cmap(self, normalized_val):
        # step 0: compute thickness and color
        max_val = 1.0
        ratio_ok = 0.7
        start_red = 0.5
        amount_green = 235 - int(235.0 * (normalized_val / (max_val * ratio_ok)) ** 4)
        amount_red = int(255 * (normalized_val / (max_val * ratio_ok)) ** 4)

        # if normalized_val < ratio_ok * max_val:
        #    amount_green = 100 - int(100. * normalized_val / (max_val * ratio_ok))
        # if normalized_val > ratio_ok:
        #    tmp = (1.0 * (normalized_val - ratio_ok) / (max_val - ratio_ok))**2
        #    amount_red += int(235. * tmp)
        # print("normalized_val {}, amount_red {}".format(normalized_val, amount_red))

        # fix to prevent pygame bug
        if amount_red < 0:
            amount_red = int(0)
        elif amount_red > 255:
            amount_red = int(255)
        if amount_green < 0:
            amount_green = int(0)
        elif amount_green > 255:
            amount_green = int(255)
        amount_red = int(amount_red)
        amount_green = int(amount_green)
        color = pygame.Color(amount_red, amount_green, 20)
        return color

    def _draw_powerlines_one_powerline(
        self, fig, l_id, pos_or, pos_ex, status, value, txt_, or_to_ex, this_col
    ):
        text_graphic = self.font.render(txt_, True, this_col)
        pos_txt = [
            int((pos_or[0] + pos_ex[0]) * 0.5),
            int((pos_or[1] + pos_ex[1]) * 0.5),
        ]
        how_center = "center|center"
        self._aligned_text(how_center, text_graphic, pos_txt)

        if not status:
            # line is disconnected
            _draw_dashed_line(self.screen, this_col, pos_or, pos_ex)
        else:
            # line is connected

            width = 1
            if value > self.rho_max:
                width = 4
            elif value > 1.0:
                width = 3
            elif value > 0.9:
                width = 2
            width += 3

            # step 1: draw the powerline with right color and thickness
            pygame.draw.line(self.screen, this_col, pos_or, pos_ex, width)

            # step 2: draw arrows indicating current flows
            _draw_arrow(
                self.screen,
                this_col,
                pos_or,
                pos_ex,
                or_to_ex,
                num_arrows=width,
                width=width,
            )

    def _aligned_text(self, pos, text_graphic, pos_text):
        if isinstance(pos_text, complex):
            pos_x = pos_text.real
            pos_y = pos_text.imag
        else:
            pos_x, pos_y = pos_text

        width = text_graphic.get_width()
        height = text_graphic.get_height()

        if pos == "center|left":
            pos_y -= height // 2
        elif pos == "up|center":
            pos_x -= width // 2
            pos_y -= height
        elif pos == "center|right":
            pos_x -= width
            pos_y -= height // 2
        elif pos == "down|center":
            pos_x -= width // 2
        elif pos == "center|center":
            pos_x -= width // 2
            pos_y -= height // 2
        self.screen.blit(text_graphic, (pos_x, pos_y))

    def _draw_loads_one_load(
        self,
        fig,
        l_id,
        pos_load,
        txt_,
        pos_end_line,
        pos_load_sub,
        how_center,
        this_col,
    ):
        width = 2
        pygame.draw.line(
            self.screen,
            this_col,
            pos_load_sub,
            (pos_end_line.real, pos_end_line.imag),
            width,
        )
        text_graphic = self.font.render(txt_, True, this_col)
        self._aligned_text(how_center, text_graphic, pos_load)

    def _draw_gens_one_gen(
        self, fig, g_id, pos_gen, txt_, pos_end_line, pos_gen_sub, how_center, this_col
    ):
        width = 2
        pygame.draw.line(
            self.screen,
            this_col,
            pos_gen_sub,
            (pos_end_line.real, pos_end_line.imag),
            width,
        )
        text_graphic = self.font.render(txt_, True, this_col)
        self._aligned_text(how_center, text_graphic, pos_gen)
        return None

    def _draw_topos_one_sub(self, fig, sub_id, buses_z, elements, bus_vect):
        colors = [pygame.Color(255, 127, 14), pygame.Color(31, 119, 180)]

        # I plot the buses
        for bus_id, z_bus in enumerate(buses_z):
            pygame.draw.circle(
                self.screen,
                colors[bus_id],
                [int(z_bus.real), int(z_bus.imag)],
                int(self.bus_radius),
                0,
            )

        # i connect every element to the proper bus with the proper color
        for el_nm, dict_el in elements.items():
            this_el_bus = bus_vect[dict_el["sub_pos"]] - 1
            if this_el_bus >= 0:
                pygame.draw.line(
                    self.screen,
                    colors[this_el_bus],
                    [int(dict_el["z"].real), int(dict_el["z"].imag)],
                    [int(buses_z[this_el_bus].real), int(buses_z[this_el_bus].imag)],
                    2,
                )
        return []



#!/usr/bin/env python

# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.


import argparse
import os
import sys

from grid2op.MakeEnv.PathUtils import DEFAULT_PATH_DATA
from grid2op.Download.DownloadDataset import main_download
from grid2op.Download.DownloadDataset import LI_VALID_ENV


def main(args=None):
    """
    DEPRECATED, use the `grid2op.make(...)`  function that will automatically download an environment
    if not present on your system already.
    """
    if args is None:
        args = download_cli()

    dataset_name = args.name
    try:
        path_data = os.path.abspath(args.path_save)
    except Exception as e:
        print(
            'Argument "--path_save" should be a valid path (directory) on your machine.'
        )
        sys.exit(1)

    try:
        main_download(dataset_name, path_data)
    except Exception as e:
        sys.exit("Aborted")


def download_cli():
    parser = argparse.ArgumentParser(
        description="Download some datasets compatible with grid2op."
    )
    parser.add_argument(
        "--path_save",
        default=DEFAULT_PATH_DATA,
        type=str,
        help="The path where the data will be downloaded.",
    )
    parser.add_argument(
        "--name",
        default="rte_case14_redisp",
        type=str,
        help="The name of the dataset (one of {} )." "".format(",".join(LI_VALID_ENV)),
    )
    args = parser.parse_args()
    return args


if __name__ == "__main__":
    args = download_cli()
    main(args)



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

import numpy as np

from grid2op.Reward.baseReward import BaseReward
from grid2op.dtypes import dt_float


class LinesReconnectedReward(BaseReward):
    """
    This reward computes a penalty
    based on the number of powerline that could have been reconnected (cooldown at 0.) but
    are still disconnected.

    Examples
    ---------
    You can use this reward in any environment with:

    .. code-block:

        import grid2op
        from grid2op.Reward import LinesReconnectedReward

        # then you create your environment with it:
        NAME_OF_THE_ENVIRONMENT = "rte_case14_realistic"
        env = grid2op.make(NAME_OF_THE_ENVIRONMENT,reward_class=LinesReconnectedReward)
        # and do a step with a "do nothing" action
        obs = env.reset()
        obs, reward, done, info = env.step(env.action_space())
        # the reward is computed with the LinesReconnectedReward class

    """

    def __init__(self, logger=None):
        BaseReward.__init__(self, logger=logger)
        self.reward_min = dt_float(0.0)
        self.reward_max = dt_float(1.0)
        self.penalty_max_at_n_lines = dt_float(2.0)

    def __call__(self, action, env, has_error, is_done, is_illegal, is_ambiguous):
        if has_error or is_illegal or is_ambiguous:
            return self.reward_min

        # Get obs from env
        obs = env.get_obs(_do_copy=False)

        # All lines ids
        lines_id = np.arange(env.n_line)
        lines_id = lines_id[obs.time_before_cooldown_line == 0]

        n_penalties = dt_float(0.0)
        for line_id in lines_id:
            # Line could be reconnected but isn't
            if obs.line_status[line_id] == False:
                n_penalties += dt_float(1.0)

        max_p = self.penalty_max_at_n_lines
        n_penalties = np.clip(n_penalties, dt_float(0.0), max_p)
        r = np.interp(
            n_penalties, [dt_float(0.0), max_p], [self.reward_max, self.reward_min]
        )
        return dt_float(r)



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

from collections import OrderedDict
import warnings
import numpy as np

from grid2op.Environment import (
    Environment,
    MultiMixEnvironment,
    BaseMultiProcessEnvironment,
)
from grid2op.Action import BaseAction, ActionSpace
from grid2op.dtypes import dt_int, dt_bool, dt_float
from grid2op.Converter.Converters import Converter
from grid2op.gym_compat.utils import GYM_AVAILABLE, GYMNASIUM_AVAILABLE


class __AuxGymActionSpace:
    """
    This class enables the conversion of the action space into a gym "space".

    Resulting action space will be a :class:`gym.spaces.Dict`.

    **NB** it is NOT recommended to use the sample of the gym action space. Please use the sampling (
    if availabe) of the original action space instead [if not available this means there is no
    implemented way to generate reliable random action]

    **Note** that gym space converted with this class should be seeded independently. It is NOT seeded
    when calling :func:`grid2op.Environment.Environment.seed`.

    .. warning::
        Depending on the presence absence of gymnasium and gym packages this class might behave differently.
        
        In grid2op we tried to maintain compatibility both with gymnasium (newest) and gym (legacy, 
        no more maintained) RL packages. The behaviour is the following:
        
        - :class:`GymActionSpace` will inherit from gymnasium if it's installed 
          (in this case it will be :class:`GymnasiumActionSpace`), otherwise it will
          inherit from gym (and will be exactly :class:`LegacyGymActionSpace`)
        - :class:`GymnasiumActionSpace` will inherit from gymnasium if it's available and never from
          from gym
        - :class:`LegacyGymActionSpace` will inherit from gym if it's available and never from
          from gymnasium
        
        See :ref:`gymnasium_gym` for more information
    
    .. note::
        A gymnasium Dict is encoded as a OrderedDict (`from collection import OrderedDict`)
        see the example section for more information.
        
    Examples
    --------
    For the "l2rpn_case14_sandbox" environment, a code using :class:`BoxGymActSpace` can look something like
    (if you want to build action "by hands"):
    
    .. code-block:: python
    
        import grid2op
        from grid2op.gym_compat import GymEnv
        import numpy as np
        env_name = "l2rpn_case14_sandbox"
        
        env = grid2op.make(env_name)
        gym_env =  GymEnv(env)
        
        obs = gym_env.reset()  # obs will be an OrderedDict (default, but you can customize it)
        
        # is equivalent to "do nothing"
        act = {}  
        obs, reward, done, truncated, info = gym_env.step(act)
        
        # you can also do a random action:
        act = gym_env.action_space.sample()
        print(gym_env.action_space.from_gym(act))
        obs, reward, done, truncated, info = gym_env.step(act)
        
        # you can chose the action you want to do (say "redispatch" for example)
        # here a random redispatch action
        act = {}
        attr_nm = "redispatch"
        act[attr_nm] = np.random.uniform(high=gym_env.action_space.spaces[attr_nm].low,
                                        low=gym_env.action_space.spaces[attr_nm].high,
                                        size=env.n_gen)
        print(gym_env.action_space.from_gym(act))
        obs, reward, done, truncated, info = gym_env.step(act)
        
    """

    # deals with the action space (it depends how it's encoded...)
    keys_grid2op_2_human = {
        "prod_p": "prod_p",
        "prod_v": "prod_v",
        "load_p": "load_p",
        "load_q": "load_q",
        "_redispatch": "redispatch",
        "_set_line_status": "set_line_status",
        "_switch_line_status": "change_line_status",
        "_set_topo_vect": "set_bus",
        "_change_bus_vect": "change_bus",
        "_hazards": "hazards",
        "_maintenance": "maintenance",
        "_storage_power": "set_storage",
        "_curtail": "curtail",
        "_raise_alarm": "raise_alarm",
        "_raise_alert": "raise_alert",
        "shunt_p": "_shunt_p",
        "shunt_q": "_shunt_q",
        "shunt_bus": "_shunt_bus",
    }
    keys_human_2_grid2op = {v: k for k, v in keys_grid2op_2_human.items()}

    def __init__(self, env, converter=None, dict_variables=None):
        """
        note: for consistency with GymObservationSpace, "action_space" here can be an environment or
        an action space or a converter
        """
        if dict_variables is None:
            dict_variables = {}
        if isinstance(
            env, (Environment, MultiMixEnvironment, BaseMultiProcessEnvironment)
        ):
            # action_space is an environment
            self.initial_act_space = env.action_space
            self._init_env = env
        elif isinstance(env, ActionSpace) and converter is None:
            warnings.warn(
                "It is now deprecated to initialize an Converter with an "
                "action space. Please use an environment instead."
            )
            self.initial_act_space = env
            self._init_env = None
        else:
            raise RuntimeError(
                "GymActionSpace must be created with an Environment of an ActionSpace (or a Converter)"
            )
        dict_ = {}
        # TODO Make sure it works well !
        if converter is not None and isinstance(converter, Converter):
            # a converter allows to ... convert the data so they have specific gym space
            self.initial_act_space = converter
            dict_ = converter.get_gym_dict(type(self))
            self.__is_converter = True
        elif converter is not None:
            raise RuntimeError(
                'Impossible to initialize a gym action space with a converter of type "{}" '
                "A converter should inherit from grid2op.Converter".format(
                    type(converter)
                )
            )
        else:
            self._fill_dict_act_space(
                dict_, self.initial_act_space, dict_variables=dict_variables
            )
            dict_ = self._fix_dict_keys(dict_)
            self.__is_converter = False
        super().__init__(dict_, dict_variables)

    def reencode_space(self, key, fun):
        """
        This function is used  to reencode the action space. For example, it can be used to scale
        the observation into values close to 0., it can also be used to encode continuous variables into
        discrete variables or the other way around etc.

        Basically, it's a tool that lets you define your own observation space (there is the same for
        the action space)

        Parameters
        ----------
        key: ``str``
            Which part of the observation space you want to study

        fun: :class:`BaseGymAttrConverter`
            Put `None` to deactivate the feature (it will be hided from the observation space)
            It can also be a `BaseGymAttrConverter`. See the example for more information.

        Returns
        -------
        self:
            The current instance, to be able to chain these calls

        Notes
        ------
        It modifies the observation space. We highly recommend to set it up at the beginning of your script
        and not to modify it afterwards

        'fun' should be deep copiable (meaning that if `copy.deepcopy(fun)` is called, then it does not crash

        If an attribute has been ignored, for example by :func`GymEnv.keep_only_obs_attr`
        or and is now present here, it will be re added in the final observation
        """
        if self._init_env is None:
            raise RuntimeError(
                "Impossible to reencode a space that has been initialized with an "
                "action space as input. Please provide a valid"
            )
        if self.__is_converter:
            raise RuntimeError(
                "Impossible to reencode a space that is a converter space."
            )

        my_dict = self.get_dict_encoding()
        if fun is not None and not isinstance(fun, type(self)._BaseGymAttrConverterType):
            raise RuntimeError(
                "Impossible to initialize a converter with a function of type {}".format(
                    type(fun)
                )
            )
        if key in self.keys_human_2_grid2op:
            key2 = self.keys_human_2_grid2op[key]
        else:
            key2 = key

        if fun is not None and not fun.is_init_space():
            if key2 in my_dict:
                fun.initialize_space(my_dict[key2])
            elif key in self.spaces:
                fun.initialize_space(self.spaces[key])
            else:
                raise RuntimeError(f"Impossible to find key {key} in your action space")
        my_dict[key2] = fun
        res = type(self)(env=self._init_env, dict_variables=my_dict)
        return res

    def _fill_dict_act_space(self, dict_, action_space, dict_variables):
        # TODO what about dict_variables !!!
        for attr_nm, sh, dt in zip(
            action_space.attr_list_vect, action_space.shape, action_space.dtype
        ):
            if sh == 0:
                # do not add "empty" (=0 dimension) arrays to gym otherwise it crashes
                continue
            my_type = None
            shape = (sh,)
            if attr_nm in dict_variables:
                # case where the user specified a dedicated encoding
                if dict_variables[attr_nm] is None:
                    # none is by default to disable this feature
                    continue
                my_type = dict_variables[attr_nm].my_space
            elif dt == dt_int:
                # discrete action space
                if attr_nm == "_set_line_status":
                    my_type = type(self)._BoxType(low=-1, high=1, shape=shape, dtype=dt)
                elif attr_nm == "_set_topo_vect":
                    my_type = type(self)._BoxType(low=-1, high=2, shape=shape, dtype=dt)
            elif dt == dt_bool:
                # boolean observation space
                my_type = self._boolean_type(sh)
                # case for all "change" action and maintenance / hazards
            else:
                # continuous observation space
                low = float("-inf")
                high = float("inf")
                shape = (sh,)
                SpaceType = type(self)._BoxType

                if attr_nm == "prod_p":
                    low = action_space.gen_pmin
                    high = action_space.gen_pmax
                    shape = None
                elif attr_nm == "prod_v":
                    # voltages can't be negative
                    low = 0.0
                elif attr_nm == "_redispatch":
                    # redispatch
                    low = -1.0 * action_space.gen_max_ramp_down
                    high = 1.0 * action_space.gen_max_ramp_up
                    low[~action_space.gen_redispatchable] = 0.0
                    high[~action_space.gen_redispatchable] = 0.0
                elif attr_nm == "_curtail":
                    # curtailment
                    low = np.zeros(action_space.n_gen, dtype=dt_float)
                    high = np.ones(action_space.n_gen, dtype=dt_float)
                    low[~action_space.gen_renewable] = -1.0
                    high[~action_space.gen_renewable] = -1.0
                elif attr_nm == "_storage_power":
                    # storage power
                    low = -1.0 * action_space.storage_max_p_prod
                    high = 1.0 * action_space.storage_max_p_absorb
                my_type = SpaceType(low=low, high=high, shape=shape, dtype=dt)

            if my_type is None:
                # if nothing has been found in the specific cases above
                my_type = self._generic_gym_space(dt, sh)

            dict_[attr_nm] = my_type

    def _fix_dict_keys(self, dict_: dict) -> dict:
        res = {}
        for k, v in dict_.items():
            res[self.keys_grid2op_2_human[k]] = v
        return res

    def from_gym(self, gymlike_action: OrderedDict) -> object:
        """
        Transform a gym-like action (such as the output of "sample()") into a grid2op action

        Parameters
        ----------
        gymlike_action: :class:`gym.spaces.dict.OrderedDict`
            The action, represented as a gym action (ordered dict)

        Returns
        -------
        An action that can be understood by the given action_space (either a grid2Op action if the
        original action space was used, or a Converter)

        """
        if self.__is_converter:
            # case where the action space comes from a converter, in this case the converter takes the
            # delegation to convert the action to openai gym
            res = self.initial_act_space.convert_action_from_gym(gymlike_action)
        else:
            # case where the action space is a "simple" action space
            res = self.initial_act_space()
            for k, v in gymlike_action.items():
                internal_k = self.keys_human_2_grid2op[k]
                if internal_k in self._keys_encoding:
                    tmp = self._keys_encoding[internal_k].gym_to_g2op(v)
                else:
                    tmp = v
                res._assign_attr_from_name(internal_k, tmp)
        return res

    def to_gym(self, action: object) -> OrderedDict:
        """
        Transform an action (non gym) into an action compatible with the gym Space.

        Parameters
        ----------
        action:
            The action (coming from grid2op or understandable by the converter)

        Returns
        -------
        gym_action:
            The same action converted as a OrderedDict (default used by gym in case of action space
            being Dict)

        """
        if self.__is_converter:
            gym_action = self.initial_act_space.convert_action_to_gym(action)
        else:
            # in that case action should be an instance of grid2op BaseAction
            assert isinstance(
                action, BaseAction
            ), "impossible to convert an action not coming from grid2op"
            # TODO this do not work in case of multiple converter,
            #  TODO this should somehow call tmp = self._keys_encoding[internal_k].g2op_to_gym(v)
            gym_action = self._base_to_gym(
                self.spaces.keys(),
                action,
                dtypes={k: self.spaces[k].dtype for k in self.spaces},
                converter=self.keys_human_2_grid2op,
            )
        return gym_action

    def close(self):
        if hasattr(self, "_init_env"):
            self._init_env = None  # this doesn't own the environment


if GYM_AVAILABLE:
    from gym.spaces import (Discrete as LegGymDiscrete,
                            Box as LegGymBox,
                            Dict as LegGymDict,
                            Space as LegGymSpace,
                            MultiBinary as LegGymMultiBinary,
                            Tuple as LegGymTuple)
    from grid2op.gym_compat.gym_space_converter import _BaseLegacyGymSpaceConverter
    from grid2op.gym_compat.base_gym_attr_converter import BaseLegacyGymAttrConverter
    LegacyGymActionSpace = type("LegacyGymActionSpace",
                                (__AuxGymActionSpace, _BaseLegacyGymSpaceConverter, ),
                                {"_DiscreteType": LegGymDiscrete,
                                 "_BoxType": LegGymBox,
                                 "_DictType": LegGymDict,
                                 "_SpaceType": LegGymSpace, 
                                 "_MultiBinaryType": LegGymMultiBinary, 
                                 "_TupleType": LegGymTuple, 
                                 "_BaseGymAttrConverterType": BaseLegacyGymAttrConverter, 
                                 "_gymnasium": False,
                                 "__module__": __name__})
    LegacyGymActionSpace.__doc__ = __AuxGymActionSpace.__doc__
    GymActionSpace = LegacyGymActionSpace
    GymActionSpace.__doc__ = __AuxGymActionSpace.__doc__
        

if GYMNASIUM_AVAILABLE:
    from gymnasium.spaces import Discrete, Box, Dict, Space, MultiBinary, Tuple
    from grid2op.gym_compat.gym_space_converter import _BaseGymnasiumSpaceConverter
    from grid2op.gym_compat.base_gym_attr_converter import BaseGymnasiumAttrConverter
    GymnasiumActionSpace = type("GymnasiumActionSpace",
                                (__AuxGymActionSpace, _BaseGymnasiumSpaceConverter, ),
                                {"_DiscreteType": Discrete,
                                 "_BoxType": Box,
                                 "_DictType": Dict,
                                 "_SpaceType": Space, 
                                 "_MultiBinaryType": MultiBinary, 
                                 "_TupleType": Tuple, 
                                 "_BaseGymAttrConverterType": BaseGymnasiumAttrConverter, 
                                 "_gymnasium": True,
                                 "__module__": __name__})
    GymnasiumActionSpace.__doc__ = __AuxGymActionSpace.__doc__
    GymActionSpace = GymnasiumActionSpace
    GymActionSpace.__doc__ = __AuxGymActionSpace.__doc__
    


__all__ = [
    "Converter",
    "ToVect",
    "IdToAct",
    "ConnectivityConverter",
    "AnalogStateConverter",
]

from grid2op.Converter.BackendConverter import BackendConverter
from grid2op.Converter.Converters import Converter
from grid2op.Converter.ToVect import ToVect
from grid2op.Converter.IdToAct import IdToAct
from grid2op.Converter.AnalogStateConverter import AnalogStateConverter
from grid2op.Converter.ConnectivityConverter import ConnectivityConverter

try:
    from grid2op.Converter.GymConverter import GymObservationSpace, GymActionSpace

    __all__.append("GymObservationSpace")
    __all__.append("GymActionSpace")
except ImportError:
    # you must install open ai gym to benefit from this converter
    pass



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

import copy
from grid2op.Reward import BaseReward
from grid2op.Action._backendAction import _BackendAction


class N1Reward(BaseReward):
    """
    This class implements the "n-1" reward, which returns the maximum flows after a powerline

    Examples
    --------

    This can be used as:

    .. code-block:: python

        import grid2op
        from grid2op.Reward import N1Reward
        L_ID = 0
        env = grid2op.make("l2rpn_case14_sandbox",
                    reward_class=N1Reward(l_id=L_ID)
                    )
        obs = env.reset()
        obs, reward, *_ = env.step(env.action_space())
        print(f"reward: {reward:.3f}")
        print("We can check that it is exactly like 'simulate' on the current step the disconnection of the same powerline")
        obs_n1, *_ = obs.simulate(env.action_space({"set_line_status": [(L_ID, -1)]}), time_step=0)
        print(f"\tmax flow after disconnection of line {L_ID}: {obs_n1.rho.max():.3f}")

    Notes
    -----
    It is also possible to use the `other_rewards` argument to simulate multiple powerline disconnections, for example:

    .. code-block:: python

        import grid2op
        from grid2op.Reward import N1Reward
        L_ID = 0
        env = grid2op.make("l2rpn_case14_sandbox",
                           other_rewards={f"line_{l_id}": N1Reward(l_id=l_id)  for l_id in [0, 1]}
                           )
        obs = env.reset()
        obs, reward, *_ = env.step(env.action_space())
        print(f"reward: {reward:.3f}")
        print("We can check that it is exactly like 'simulate' on the current step the disconnection of the same powerline")
        obs_n1, *_ = obs.simulate(env.action_space({"set_line_status": [(L_ID, -1)]}), time_step=0)
        print(f"\tmax flow after disconnection of line {L_ID}: {obs_n1.rho.max():.3f}")

    """

    def __init__(self, l_id=0, logger=None):
        BaseReward.__init__(self, logger=logger)
        self._backend = None
        self._backend_action = None
        self.l_id = l_id

    def initialize(self, env):
        self._backend = env.backend.copy()
        bk_act_cls = _BackendAction.init_grid(env.backend)
        self._backend_action = bk_act_cls()

    def __call__(self, action, env, has_error, is_done, is_illegal, is_ambiguous):
        if is_done:
            return self.reward_min
        self._backend_action.reset()
        act = env.backend.get_action_to_set()
        th_lim = env.get_thermal_limit()
        th_lim[th_lim <= 1] = 1  # assign 1 for the thermal limit
        this_n1 = copy.deepcopy(act)
        self._backend_action += this_n1
            
        self._backend.apply_action(self._backend_action)
        self._backend._disconnect_line(self.l_id)
        div_exc_ = None
        try:
            # TODO there is a bug in lightsimbackend that make it crash instead of diverging
            conv, div_exc_ = self._backend.runpf()
        except Exception as exc_:
            conv = False
            div_exc_ = exc_

        if conv:
            flow = self._backend.get_line_flow()
            res = (flow / th_lim).max()
        else:
            self.logger.info(f"Divergence of the backend at step {env.nb_time_step} for N1Reward with error `{div_exc_}`")
            res = -1
        return res

    def close(self):
        self._backend.close()
        del self._backend
        self._backend = None



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.
from multiprocessing import Process, Pipe
import numpy as np

from grid2op.dtypes import dt_int
from grid2op.Exceptions import Grid2OpException, MultiEnvException
from grid2op.Space import GridObjects
from grid2op.Environment.baseMultiProcessEnv import BaseMultiProcessEnvironment
from grid2op.Action import BaseAction


class MultiEnvMultiProcess(BaseMultiProcessEnvironment):
    """
    This class allows to evaluate a single agent instance on multiple environments running in parrallel.

    It is a kind of :class:`BaseMultiProcessEnvironment`. For more information you can consult the
    documentation of this parent class. This class allows to interact at the same time with different copy of
    possibly different environments in parallel

    Attributes
    -----------
    envs: `list:grid2op.Environment.Environment`
        Al list of environments for which the evaluation will be made in parallel.

    nb_envs: ``list:int``
        Number of parallel underlying environment that will be handled.
        MUST be the same length as the parameter `envs`.
        The total number of subprocesses will be the sum of this list.


    Examples
    --------
    This class can be used as:

    .. code-block:: python

        import grid2op
        from grid2op.Environment import MultiEnvMultiProcess
        env0 = grid2op.make()  # create an environment
        env1 = grid2op.make()  # create a second environment, that can be similar, or not
        # it is recommended to filter or create the environment with different parameters, otherwise this class
        # is of little interest
        envs = [env0, env1]  # list of all environments created
        nb_envs = [1, 7]  # number of "copies" of each environment that will be made.
        # in this case the first one will be copied only once, and the second one 7 times.
        # the total number of environments used in the multi env will be the sum(nb_envs), here 8.

        multi_env = MultiEnvMultiProcess(envs=envs, nb_envs=nb_envs)
        # and now you can use it like any other grid2op environment (almost)
        observations = multi_env.reset()

    """

    def __init__(self, envs, nb_envs, obs_as_class=True, return_info=True, logger=None):
        try:
            nb_envs = np.array(nb_envs)
            nb_envs = nb_envs.astype(dt_int)
        except Exception as exc_:
            raise MultiEnvException(
                '"nb_envs" argument should be a list of integers. We could not '
                'convert it to such with error "{}"'.format(exc_)
            )

        if (nb_envs < 0).any():
            raise MultiEnvException(
                'You ask to perform "{}" copy of an environment. This is a negative '
                'integer. I cannot do that. Please make sure "nb_envs" argument '
                "is all made of strictly positive integers and not {}."
                "".format(np.min(nb_envs), nb_envs)
            )
        if (nb_envs == 0).any():
            raise MultiEnvException(
                "You ask to perform 0 copy of an environment. This is not supported at "
                'the moment. Please make sure "nb_envs" argument '
                "is all made of strictly positive integers and not {}."
                "".format(nb_envs)
            )

        all_envs = []
        for e, n in enumerate(nb_envs):
            all_envs += [envs[e] for _ in range(n)]
        super().__init__(
            all_envs,
            obs_as_class=obs_as_class,
            return_info=return_info,
            logger=logger.getChild("MultiEnvMultiProcess")
            if logger is not None
            else None,
        )


if __name__ == "__main__":
    from tqdm import tqdm
    from grid2op import make
    from grid2op.Agent import DoNothingAgent

    nb_env = [2, 2, 1, 1, 2]  # change that to adapt to your system
    NB_STEP = 100  # number of step for each environment

    env = make()
    env.seed(42)
    envs = [env, env, env, env, env]

    agent = DoNothingAgent(env.action_space)
    multi_envs = MultiEnvMultiProcess(envs, nb_env)

    obs = multi_envs.reset()
    rews = [env.reward_range[0] for i in range(multi_envs.nb_env)]
    dones = [False for i in range(multi_envs.nb_env)]

    total_reward = 0.0
    for i in tqdm(range(NB_STEP)):
        acts = [None for _ in range(multi_envs.nb_env)]
        for env_act_id in range(multi_envs.nb_env):
            acts[env_act_id] = agent.act(
                obs[env_act_id], rews[env_act_id], dones[env_act_id]
            )
        obs, rews, dones, infos = multi_envs.step(acts)
        total_reward += np.sum(rews)
        len(rews)

    multi_envs.close()

    ob = env.reset()
    rew = env.reward_range[0]
    done = False
    total_reward_single = 0
    for i in tqdm(range(NB_STEP)):
        act = agent.act(ob, rew, done)
        ob, rew, done, info = env.step(act)
        if done:
            ob = env.reset()
        total_reward_single += np.sum(rew)
    env.close()
    print("total_reward mluti_env: {}".format(total_reward))
    print("total_reward single env: {}".format(total_reward_single))



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

import numpy as np
from grid2op.Exceptions import OpponentError


class BaseActionBudget:
    """
    INTERNAL

    .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\

    This is the base class representing the action budget.
    It makes sure the opponent uses the correct type of "action", and  compute the bugdet associated to it.

    """

    def __init__(self, action_space):
        self.action_space = action_space

    def __call__(self, attack):
        """
        INTERNAL
        
        .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\

        This function takes an attack as input and compute the cost associated to it.

        **NB** The cost of a "None" attack is necessarily 0 !

        Parameters
        ----------
        attack: :class:`ŋrid2op.BaseAction.BaseAction`
            The attack performed by the opponent

        Returns
        -------
        cost: the cost of the action performed by the opponent.

        """
        if attack is None:
            return 0

        if not isinstance(attack, self.action_space.actionClass):
            raise OpponentError(
                'Attempt to use an attack of type "{}" which is not a instance of "{}", '
                "the type of action the opponent was supposed to use."
                "".format(type(attack), self.action_space.actionClass)
            )
        aff_lines, aff_subs = attack.get_topological_impact()
        cost = aff_lines.sum() + aff_subs.sum()
        return cost



# Copyright (c) 2023, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.


class HighResSimCounter:
    """This classes helps to count the total number of call to "high fidelity simulator"
    the agent made.
    """
    def __init__(self) -> None:
        self.__nb_highres_called = 0
        
    def __iadd__(self, other):
        self.__nb_highres_called += int(other)
        return self
        
    def add_one(self):
        self.__nb_highres_called += 1
    
    @property
    def nb_highres_called(self):
        return self.__nb_highres_called
    


# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

import numpy as np

from grid2op.Environment.baseMultiProcessEnv import BaseMultiProcessEnvironment


class SingleEnvMultiProcess(BaseMultiProcessEnvironment):
    """
    This class allows to evaluate a single agent instance on multiple environments running in parallel.

    It is a kind of :class:`BaseMultiProcessEnvironment`. For more information you can consult the
    documentation of this parent class. It allows to interact at the same time with different copy of the
    (same) environment in parallel

    Attributes
    -----------
    env: `list::grid2op.Environment.Environment`
        Al list of environments for which the evaluation will be made in parallel.

    nb_env: ``int``
        Number of parallel underlying environment that will be handled. It is also the size of the list of actions
        that need to be provided in :func:`MultiEnvironment.step` and the return sizes of the list of this
        same function.

    Examples
    --------

    An example on how you can best leverage this class is given in the getting_started notebooks. Another simple
    example is:

    .. code-block:: python

        from grid2op.BaseAgent import DoNothingAgent
        from grid2op.MakeEnv import make
        from grid2op.Environment import SingleEnvMultiProcess

        # create a simple environment
        env = make()
        # number of parrallel environment
        nb_env = 2  # change that to adapt to your system
        NB_STEP = 100  # number of step for each environment

        # create a simple agent
        agent = DoNothingAgent(env.action_space)

        # create the multi environment class
        multi_envs = SingleEnvMultiProcess(env=env, nb_env=nb_env)

        # making is usable
        obs = multi_envs.reset()
        rews = [env.reward_range[0] for i in range(nb_env)]
        dones = [False for i in range(nb_env)]

        # performs the appropriated steps
        for i in range(NB_STEP):
            acts = [None for _ in range(nb_env)]
            for env_act_id in range(nb_env):
                acts[env_act_id] = agent.act(obs[env_act_id], rews[env_act_id], dones[env_act_id])
            obs, rews, dones, infos = multi_envs.step(acts)

            # DO SOMETHING WITH THE AGENT IF YOU WANT

        # close the environments
        multi_envs.close()
        # close the initial environment
        env.close()

    """

    def __init__(self, env, nb_env, obs_as_class=True, return_info=True, logger=None):
        envs = [env for _ in range(nb_env)]
        super().__init__(
            envs,
            obs_as_class=obs_as_class,
            return_info=return_info,
            logger=logger.getChild("SingleEnvMultiProcess")
            if logger is not None
            else None,
        )


if __name__ == "__main__":
    from tqdm import tqdm
    from grid2op import make
    from grid2op.Agent import DoNothingAgent

    nb_env = 8  # change that to adapt to your system
    NB_STEP = 100  # number of step for each environment

    env = make()
    env.seed(42)

    agent = DoNothingAgent(env.action_space)
    multi_envs = SingleEnvMultiProcess(env, nb_env)

    obs = multi_envs.reset()
    rews = [env.reward_range[0] for i in range(nb_env)]
    dones = [False for i in range(nb_env)]

    total_reward = 0.0
    for i in tqdm(range(NB_STEP)):
        acts = [None for _ in range(nb_env)]
        for env_act_id in range(nb_env):
            acts[env_act_id] = agent.act(
                obs[env_act_id], rews[env_act_id], dones[env_act_id]
            )
        obs, rews, dones, infos = multi_envs.step(acts)
        total_reward += np.sum(rews)
        len(rews)

    multi_envs.close()

    ob = env.reset()
    rew = env.reward_range[0]
    done = False
    total_reward_single = 0
    for i in tqdm(range(NB_STEP)):
        act = agent.act(ob, rew, done)
        ob, rew, done, info = env.step(act)
        if done:
            ob = env.reset()
        total_reward_single += np.sum(rew)
    env.close()
    print("total_reward mluti_env: {}".format(total_reward))
    print("total_reward single env: {}".format(total_reward_single))



from grid2op.Action import PowerlineSetAction
from grid2op.Reward import RedispReward
from grid2op.Rules import DefaultRules
from grid2op.Chronics import Multifolder
from grid2op.Chronics import GridStateFromFileWithForecasts
from grid2op.Backend import PandaPowerBackend
from grid2op.Opponent import WeightedRandomOpponent, BaseActionBudget


try:
    from grid2op.l2rpn_utils import ActionNeurips2020, ObservationNeurips2020
except ImportError:
    from grid2op.Action import TopologyAndDispatchAction
    from grid2op.Observation import CompleteObservation
    import warnings
    warnings.warn("The grid2op version you are trying to use is too old for this environment. Please upgrade it.")
    ActionNeurips2020 = TopologyAndDispatchAction
    ObservationNeurips2020 = CompleteObservation
    
    
lines_attacked = [
    "62_58_180",
    "62_63_160",
    "48_50_136",
    "48_53_141",
    "41_48_131",
    "39_41_121",
    "43_44_125",
    "44_45_126",
    "34_35_110",
    "54_58_154",
]
rho_normalization = [0.45, 0.45, 0.6, 0.35, 0.3, 0.2, 0.55, 0.3, 0.45, 0.55]
opponent_attack_cooldown = 12 * 24  # 24 hours, 1 hour being 12 time steps
opponent_attack_duration = 12 * 4  # 4 hours
opponent_budget_per_ts = (
    0.16667  # opponent_attack_duration / opponent_attack_cooldown + epsilon
)
opponent_init_budget = 144.0  # no need to attack straightfully, it can attack starting at midday the first day
config = {
    "backend": PandaPowerBackend,
    "action_class": ActionNeurips2020,
    "observation_class": ObservationNeurips2020,
    "reward_class": RedispReward,
    "gamerules_class": DefaultRules,
    "chronics_class": Multifolder,
    "grid_value_class": GridStateFromFileWithForecasts,
    "volagecontroler_class": None,
    "names_chronics_to_grid": None,
    "thermal_limits": [
        60.9,
        231.9,
        272.6,
        212.8,
        749.2,
        332.4,
        348.0,
        414.4,
        310.1,
        371.4,
        401.2,
        124.3,
        298.5,
        86.4,
        213.9,
        160.8,
        112.2,
        291.4,
        489.0,
        489.0,
        124.6,
        196.7,
        191.9,
        238.4,
        174.2,
        105.6,
        143.7,
        293.4,
        288.9,
        107.7,
        415.5,
        148.2,
        124.2,
        154.4,
        85.9,
        106.5,
        142.0,
        124.0,
        130.2,
        86.2,
        278.1,
        182.0,
        592.1,
        173.1,
        249.8,
        441.0,
        344.2,
        722.8,
        494.6,
        494.6,
        196.7,
        151.8,
        263.4,
        364.1,
        327.0,
        370.5,
        441.0,
        300.3,
        656.2,
    ],
    "opponent_attack_cooldown": opponent_attack_cooldown,
    "opponent_attack_duration": opponent_attack_duration,
    "opponent_budget_per_ts": opponent_budget_per_ts,
    "opponent_init_budget": opponent_init_budget,
    "opponent_action_class": PowerlineSetAction,
    "opponent_class": WeightedRandomOpponent,
    "opponent_budget_class": BaseActionBudget,
    "kwargs_opponent": {
        "lines_attacked": lines_attacked,
        "rho_normalization": rho_normalization,
        "attack_period": opponent_attack_cooldown,
    },
}



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

import copy
import numpy as np

from grid2op.dtypes import dt_float
from grid2op.gym_compat.utils import GYM_AVAILABLE, GYMNASIUM_AVAILABLE


class __AuxScalerAttrConverter:
    """
    This is a scaler that transforms a initial gym space `init_space` into its scale version.

    It can be use to scale the observation by substracting the mean and dividing by the variance for
    example.

    TODO work in progress !

    Need help if you can :-)

    .. warning::
        Depending on the presence absence of gymnasium and gym packages this class might behave differently.
        
        In grid2op we tried to maintain compatibility both with gymnasium (newest) and gym (legacy, 
        no more maintained) RL packages. The behaviour is the following:
        
        - :class:`ScalerAttrConverter` will inherit from gymnasium if it's installed 
          (in this case it will be :class:`ScalerAttrConverterGymnasium`), otherwise it will
          inherit from gym (and will be exactly :class:`ScalerAttrConverterLegacyGym`)
        - :class:`ScalerAttrConverterGymnasium` will inherit from gymnasium if it's available and never from
          from gym
        - :class:`ScalerAttrConverterLegacyGym` will inherit from gym if it's available and never from
          from gymnasium
        
        See :ref:`gymnasium_gym` for more information
        
    """

    def __init__(self, substract, divide, dtype=None, init_space=None):
        super().__init__(
            g2op_to_gym=None, gym_to_g2op=None, space=None
        )  # super should be from type BaseGymAttrConverter
        self._substract = np.array(substract)
        self._divide = np.array(divide)
        self.dtype = dtype if dtype is not None else dt_float
        if init_space is not None:
            self.initialize_space(init_space)

    def initialize_space(self, init_space):
        if self._is_init_space:
            return
        if not isinstance(init_space, type(self)._BoxType):
            raise RuntimeError(
                "Impossible to scale a converter if this one is not from type space.Box"
            )

        tmp_space = copy.deepcopy(init_space)
        # properly change the low / high value
        low_tmp = self.scale(tmp_space.low)
        high_tmp = self.scale(tmp_space.high)
        low_ = np.minimum(high_tmp, low_tmp)
        high_ = np.maximum(high_tmp, low_tmp)
        tmp_space.low[:] = low_
        tmp_space.high[:] = high_

        if self.dtype is not None:
            tmp_space.dtype = np.dtype(self.dtype)
            tmp_space.low = tmp_space.low.astype(self.dtype)
            tmp_space.high = tmp_space.high.astype(self.dtype)
        self.base_initialize(
            space=tmp_space, g2op_to_gym=self.scale, gym_to_g2op=self.unscale
        )
        self.dtype = self.my_space.dtype
        self._substract = self._substract.astype(self.dtype)
        self._divide = self._divide.astype(self.dtype)
        self._is_init_space = True

    def scale(self, vect):
        tmp = vect.astype(self.dtype)
        tmp = (tmp - self._substract) / self._divide
        return tmp

    def unscale(self, vect):
        tmp = vect * self._divide + self._substract
        return tmp

    def close(self):
        pass


if GYM_AVAILABLE:
    from gym.spaces import Box as LegacyGymBox
    from grid2op.gym_compat.base_gym_attr_converter import BaseLegacyGymAttrConverter
    ScalerAttrConverterLegacyGym = type("ScalerAttrConverterLegacyGym",
                                        (__AuxScalerAttrConverter, BaseLegacyGymAttrConverter, ),
                                        {"_gymnasium": False,
                                         "_BoxType": LegacyGymBox,
                                         "__module__": __name__})
    ScalerAttrConverterLegacyGym.__doc__ = __AuxScalerAttrConverter.__doc__
    ScalerAttrConverter = ScalerAttrConverterLegacyGym
    ScalerAttrConverter.__doc__ = __AuxScalerAttrConverter.__doc__
        

if GYMNASIUM_AVAILABLE:
    from gymnasium.spaces import Box
    from grid2op.gym_compat.base_gym_attr_converter import BaseGymnasiumAttrConverter
    ScalerAttrConverterGymnasium = type("ScalerAttrConverterGymnasium",
                                     (__AuxScalerAttrConverter, BaseGymnasiumAttrConverter, ),
                                     {"_gymnasium": True,
                                      "_BoxType": Box,
                                      "__module__": __name__})
    ScalerAttrConverterGymnasium.__doc__ = __AuxScalerAttrConverter.__doc__
    ScalerAttrConverter = ScalerAttrConverterGymnasium
    ScalerAttrConverter.__doc__ = __AuxScalerAttrConverter.__doc__



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.
from grid2op.Action import ActionSpace


class Converter(ActionSpace):
    """
    This Base class should be use to implement any converter. If for some reasons
    """

    def __init__(self, action_space):
        ActionSpace.__init__(
            self, action_space, action_space.legal_action, action_space.subtype
        )
        self.space_prng = action_space.space_prng
        self.seed_used = action_space.seed_used

    def init_converter(self, **kwargs):
        pass

    def convert_obs(self, obs):
        """
        This function is used to convert an observation into something that is easier to manipulate.

        Parameters
        ----------
        obs: :class:`grid2op.Observation.Observation`
            The input observation.

        Returns
        -------

        transformed_obs: ``object``
            An different representation of the input observation, typically represented as a 1d vector that can be
            processed by a neural networks.

        """
        transformed_obs = obs
        return transformed_obs

    def convert_act(self, encoded_act):
        """
        This function will transform the action, encoded somehow (for example identified by an id, represented by
        an integer) to a valid actions that can be processed by the environment.

        Parameters
        ----------
        encoded_act: ``object``
            Representation of an action, as a vector or an integer etc.

        Returns
        -------
        regular_act: :class:`grid2op.Action.Action`
            The action corresponding to the `encoded_action` above converted into a format that can be processed
            by the environment.

        """
        regular_act = encoded_act
        return regular_act

    def get_gym_dict(self, cls_gym):
        """
        To convert this space into a open ai gym space. This function returns a dictionnary used
        to initialize such a converter.

        It should not be used directly. Prefer to use the :class:`grid2op.Converter.GymConverter`

        cls_gym represents either :class:`grid2op.gym_compat.LegacyGymActionSpace` or
        :class:`grid2op.gym_compat.GymnasiumActionSpace`
        """
        raise NotImplementedError(
            'Impossible to convert the converter "{}" automatically '
            "into a gym space (or gym is not installed on your machine)."
            "".format(self)
        )

    def convert_action_from_gym(self, gymlike_action):
        """
        Convert the action (represented as a gym object, in fact an ordered dict) as an action
        compatible with this converter.

        This is not compatible with all converters and you need to install gym for it to work.

        Parameters
        ----------
        gymlike_action:
            the action to be converted to an action compatible with the action space representation

        Returns
        -------
        res:
            The action converted to be understandable by this converter.

        Examples
        ---------
        Here is an example on how to use this feature with the :class:`grid2op.Converter.IdToAct`
        converter (imports are not shown here).

        .. code-block:: python

            # create the environment
            env = grid2op.make()

            # create the converter
            converter = IdToAct(env.action_space)

            # create the gym action space
            gym_action_space = GymObservationSpace(action_space=converter)

            gym_action = gym_action_space.sample()
            converter_action = converter.convert_action_from_gym(gym_action)  # this represents the same action
            grid2op_action = converter.convert_act(converter_action)

        """
        raise NotImplementedError(
            "Impossible to convert the gym-like action automatically "
            'into the converter representation for "{}" '
            "".format(self)
        )

    def convert_action_to_gym(self, action):
        """
        Convert the action (compatible with this converter) into a "gym action" (ie an OrderedDict)

        This is not compatible with all converters and you need to install gym for it to work.

        Parameters
        ----------
        action:
            the action to be converted to an action compatible with the action space representation

        Returns
        -------
        res:
            The action converted to a "gym" model (can be used by a machine learning model)

        Examples
        ---------
        Here is an example on how to use this feature with the :class:`grid2op.Converter.IdToAct`
        converter (imports are not shown here).

        .. code-block:: python

            # create the environment
            env = grid2op.make()

            # create the converter
            converter = IdToAct(env.action_space)

            # create the gym action space
            gym_action_space = GymObservationSpace(action_space=converter)

            converter_action = converter.sample()
            gym_action = converter.to_gym(converter_action)  # this represents the same action

        """
        raise NotImplementedError(
            "Impossible to convert the gym-like action automatically "
            'into the converter representation for "{}" '
            "".format(self)
        )



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

import os
import json

from grid2op.MakeEnv.Make import _list_available_remote_env_aux
import grid2op.MakeEnv.PathUtils
from grid2op.MakeEnv.PathUtils import DEFAULT_PATH_CONFIG, KEY_DATA_PATH
from grid2op.Exceptions import Grid2OpException


def list_available_remote_env():
    """
    This function returns the list of available environments. It returns all the environments, you might already
    have downloaded some, they will be listed here.

    Returns
    -------
    res: ``list``
        a sorted list of available to environments that can be downloaded.

    Examples
    ---------
    A usage example is

    .. code-block:: python

        import grid2op
        li = grid2op.list_available_remote_env()
        li_fmt = '\\n * '.join(li)
        print(f"The available environments are: \\n * {li_fmt}")

    """
    avail_datasets_json = _list_available_remote_env_aux()
    return sorted(avail_datasets_json.keys())


def list_available_local_env():
    """
    This function returns the environment that are available locally. It does not return the environments that
    are included in the package.

    Returns
    -------
    res: ``list``
        a sorted list of available environments locally.

    Examples
    ---------

    .. code-block:: python

        import grid2op
        li = grid2op.list_available_local_env()
        li_fmt = '\\n + '.join(li)
        print(f"The locally available environments (without downloading anything) are: \\n * {li_fmt}")

    """
    res = []
    if not os.path.exists(grid2op.MakeEnv.PathUtils.DEFAULT_PATH_DATA):
        return res

    for el in os.listdir(grid2op.MakeEnv.PathUtils.DEFAULT_PATH_DATA):
        tmp_dir = os.path.join(grid2op.MakeEnv.PathUtils.DEFAULT_PATH_DATA, el)
        if (
            os.path.exists(os.path.join(tmp_dir, "config.py"))
            and os.path.exists(os.path.join(tmp_dir, "grid.json"))
        ) or os.path.exists(os.path.join(tmp_dir, ".multimix")):
            res.append(el)
    return res


def list_available_test_env():
    """
    This functions list the environment available through "grid2op.make(..., test=True)", which are the environment
    used for testing purpose, but available without the need to download any data.

    The "test" environment are provided with the grid2op package.

    Returns
    -------
    res: ``list``
        a sorted list of available environments for testing / illustration purpose.

    Examples
    ---------

    .. code-block:: python

        import grid2op
        li = grid2op.list_available_test_env()

        env = grid2op.make(li[0], test=True)

    """
    from grid2op.MakeEnv.Make import TEST_DEV_ENVS
    import re

    res = sorted(
        [
            el
            for el in TEST_DEV_ENVS.keys()
            if re.match("(^rte_.*)|(^l2rpn_.*)|(^educ_.*)", el) is not None
        ]
    )
    return res


def get_current_local_dir():
    """
    This function allows you to get the directory in which grid2op will download the datasets. This path can
    be modified with the ".grid2opconfig.json" file.

    Returns
    -------
    res: ``str``
        The current path were data are downloaded in.

    Examples
    ---------

    .. code-block:: python

        import grid2op
        print(f"Data about grid2op downloaded environments are stored in: \"{grid2op.get_current_local_dir()}\"")

    """
    return os.path.abspath(grid2op.MakeEnv.PathUtils.DEFAULT_PATH_DATA)


def change_local_dir(new_path):
    """
    This function will change the path were datasets are read to / from.

    The previous datasets will be left in the previous configuration folder and will not be accessible by other
    grid2op function such as "make" for example.

    Parameters
    ----------
    new_path: ``str``
        The new path in which to download the datasets.

    Examples
    ---------
    To set the download path, and the path where grid2op will look for available local environment you can:

    .. code-block:: python

        import grid2op
        local_dir = ...  # should be a valid path on your machine
        grid2op.change_local_dir(local_dir)

        # check it has worked:
        print(f"Data about grid2op downloaded environments are now stored in: \"{grid2op.get_current_local_dir()}\"")

    """

    try:
        new_path = str(new_path)
    except:
        raise Grid2OpException(
            'The new path should be convertible to str. It is currently "{}"'.format(
                new_path
            )
        )

    root_dir = os.path.split(new_path)[0]
    if not os.path.exists(root_dir):
        raise Grid2OpException(
            'Data cannot be stored in "{}" as the base path of this directory ("{}") does '
            "not exists.".format(new_path, root_dir)
        )

    if not os.path.isdir(new_path):
        raise Grid2OpException(
            'Data cannot be stored in "{}" as it is a file and not a directory.'.format(
                new_path
            )
        )

    newconfig = {}

    if os.path.exists(DEFAULT_PATH_CONFIG):
        try:
            with open(DEFAULT_PATH_CONFIG, "r", encoding="utf-8") as f:
                newconfig = json.load(f)
        except:
            raise Grid2OpException(
                'Impossible to read the grid2op configuration files "{}". Make sure it is a '
                'valid json encoded with "utf-8" encoding.'.format(DEFAULT_PATH_CONFIG)
            )

    newconfig[KEY_DATA_PATH] = new_path

    try:
        with open(DEFAULT_PATH_CONFIG, "w", encoding="utf-8") as f:
            json.dump(fp=f, obj=newconfig, sort_keys=True, indent=4)
    except:
        raise Grid2OpException(
            'Impossible to write the grid2op configuration files "{}". Make sure you have '
            "writing access to it.".format(DEFAULT_PATH_CONFIG)
        )

    grid2op.MakeEnv.PathUtils.DEFAULT_PATH_DATA = new_path



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

import numpy as np
from grid2op.Agent import BaseAgent
from grid2op.Observation import BaseObservation
from grid2op.Action import ActionSpace
from grid2op.Exceptions import AgentError


class RecoPowerlinePerArea(BaseAgent):
    """This class acts like the :class:`RecoPowerlineAgent` but it is able
    to reconnect multiple lines at the same steps (one line per area).
    
    The "areas" are defined by a list of list of substation id provided as input.

    Of course the area you provide to the agent should be the same as the areas
    used in the rules of the game. Otherwise, the agent might try to reconnect
    two powerline "in the same area for the environment" which of course will
    lead to an illegal action.

    You can use it like:
    
    .. code-block::
    
        import grid2op
        from grid2op.Agent import RecoPowerlinePerArea

        env_name = "l2rpn_idf_2023" # (or any other env name supporting the feature)
        env = grid2op.make(env_name)
        agent = RecoPowerlinePerArea(env.action_space, env._game_rules.legal_action.substations_id_by_area)
        

    """
    def __init__(self, action_space: ActionSpace, areas_by_sub_id: dict):
        super().__init__(action_space)
        self.lines_to_area_id = np.zeros(type(action_space).n_line, dtype=int) - 1
        for aread_id, (area_nm, sub_this_area) in enumerate(areas_by_sub_id.items()):
            for line_id, subor_id in enumerate(type(action_space).line_or_to_subid):
                if subor_id in sub_this_area:
                    self.lines_to_area_id[line_id] = aread_id
        if (self.lines_to_area_id == -1).any():
            raise AgentError("some powerline have no area id")
        self.nb_area = len(areas_by_sub_id)
    
    def act(self, observation: BaseObservation, reward: float, done : bool=False):
        line_stat_s = observation.line_status
        cooldown = observation.time_before_cooldown_line
        can_be_reco = ~line_stat_s & (cooldown == 0)
        if not can_be_reco.any():
            # no line to reconnect
            return self.action_space()
        area_used = np.full(self.nb_area, fill_value=False, dtype=bool)
        reco_ids = []
        for l_id in np.where(can_be_reco)[0]:
            if not area_used[self.lines_to_area_id[l_id]]:
                reco_ids.append(l_id)
                area_used[self.lines_to_area_id[l_id]] = True
        res = self.action_space({"set_line_status": [(l_id, +1) for l_id in reco_ids]})
        return res
